import logging
import os

import ibmsecurity.isam.web.api_access_control.resources.instances
import ibmsecurity.isam.web.api_access_control.resources.servers
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/resource/instance"
requires_modules = ["wga"]
requires_version = "9.0.7"


def get_all(isamAppliance, instance_name, resource_server_name, check_mode=False, force=False):
    """
    Retrieve a list of all API Access Control Resources
    """
    return isamAppliance.invoke_get("Retrieve a list of all API Access Control Resources",
                                    "{0}/{1}/server{2}/resource".format(uri, instance_name, resource_server_name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, instance_name, resource_server_name, resource_name, method,
        server_type="standard", check_mode=False, force=False):
    """
    Retrieve a single API Access Control Resource
    """

    if tools.version_compare(isamAppliance.facts["version"], "10.0.0") < 0:
        url = "{0}/{1}/server{2}/resource/{3}{4}?server_type={5}".format(uri, instance_name, resource_server_name,
                                                                         method, resource_name, server_type)
    else:
        url = "{0}/{1}/server{2}/resource/{3}{2}{4}?server_type={5}".format(uri, instance_name, resource_server_name,
                                                                            method, resource_name, server_type)

    return isamAppliance.invoke_get("Retrieve a single API Access Control Resource",
                                    url, requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, instance_name, resource_server_name, method, path, policy, server_type="standard",
        cors_policy='', static_response_headers=[], url_aliases=[], name=None, rate_limiting_policy=None,
        documentation=None, check_mode=False, force=False):
    """
    Creating a new API Access Control Resource
    """
    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    resource_exist, warnings = _check_resource_exist(isamAppliance, instance_name, resource_server_name, method, path)

    if force is True or resource_exist is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            json_data = {
                'method': method,
                'policy': policy,
                'server_type': server_type,
                'cors_policy': cors_policy,
                'static_response_headers': static_response_headers,
                'url_aliases': url_aliases
            }

            if tools.version_compare(isamAppliance.facts["version"], "10.0.0") < 0:
                json_data['path'] = path
            else:
                json_data['path'] = "{0}{1}".format(resource_server_name, path)

            if name is not None:
                json_data['name'] = name

            if rate_limiting_policy is not None:
                json_data['rate_limiting_policy'] = rate_limiting_policy

            if documentation is not None:
                json_data['documentation'] = documentation

            return isamAppliance.invoke_post(
                "Creating a new API Access Control Resource",
                "{0}/{1}/server{2}/resource".format(uri, instance_name, resource_server_name),
                json_data, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, instance_name, resource_server_name, method, path, policy, server_type="standard",
           cors_policy='', static_response_headers=[], url_aliases=[], name=None, rate_limiting_policy=None,
           documentation=None, check_mode=False, force=False):
    """
    Updating an existing API Access Control Resource
    """
    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)
            else:
                resource_exist, warnings = _check_resource_exist(isamAppliance, instance_name, resource_server_name,
                                                                 method, path)
                if resource_exist is False:
                    warnings.append("The specified resource does not exist")
                    return isamAppliance.create_return_object(warnings=warnings)

    update_required, warnings, json_data = _check_resource_content(isamAppliance=isamAppliance,
                                                                   instance_name=instance_name,
                                                                   resource_server_name=resource_server_name,
                                                                   method=method, path=path, policy=policy,
                                                                   cors_policy=cors_policy, name=name,
                                                                   static_response_headers=static_response_headers,
                                                                   rate_limiting_policy=rate_limiting_policy,
                                                                   url_aliases=url_aliases, documentation=documentation)
    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if tools.version_compare(isamAppliance.facts["version"], "10.0.0") < 0:
                url = "{0}/{1}/server{2}/resource/{3}{4}?server_type={5}".format(uri, instance_name,
                                                                                 resource_server_name,
                                                                                 method, path, server_type)
            else:
                url = "{0}/{1}/server{2}/resource/{3}{2}{4}?server_type={5}".format(uri, instance_name,
                                                                                    resource_server_name,
                                                                                    method, path, server_type)

            return isamAppliance.invoke_put(
                "Updating an existing API Access Control Resource",
                url, json_data, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, instance_name, resource_server_name, method, path, policy, server_type="standard",
        cors_policy='', static_response_headers=[], url_aliases=[], name=None, rate_limiting_policy=None,
        documentation=None, check_mode=False, force=False):
    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    resource_exist, warnings = _check_resource_exist(isamAppliance, instance_name, resource_server_name,
                                                     method, path)
    if resource_exist is True:
        return update(isamAppliance=isamAppliance, instance_name=instance_name,
                      resource_server_name=resource_server_name, method=method, path=path, policy=policy,
                      server_type=server_type, cors_policy=cors_policy, static_response_headers=static_response_headers,
                      url_aliases=url_aliases, name=name, rate_limiting_policy=rate_limiting_policy,
                      documentation=documentation, check_mode=check_mode, force=force)
    else:
        return add(isamAppliance=isamAppliance, instance_name=instance_name,
                   resource_server_name=resource_server_name, method=method, path=path, policy=policy,
                   server_type=server_type, cors_policy=cors_policy, static_response_headers=static_response_headers,
                   url_aliases=url_aliases, name=name, rate_limiting_policy=rate_limiting_policy,
                   documentation=documentation, check_mode=check_mode, force=force)


def delete(isamAppliance, instance_name, resource_server_name, method, path, server_type='standard',
           check_mode=False, force=False):
    """
    Delete an existing API Access Control Resource
    """

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    resource_exist, warnings = _check_resource_exist(isamAppliance, instance_name, resource_server_name, method, path)

    if force is True or resource_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if tools.version_compare(isamAppliance.facts["version"], "10.0.0") < 0:
                url = "{0}/{1}/server{2}/resource/{3}{4}?server_type={5}".format(uri, instance_name,
                                                                                 resource_server_name,
                                                                                 method, path, server_type)
            else:
                url = "{0}/{1}/server{2}/resource/{3}{2}{4}?server_type={5}".format(uri, instance_name,
                                                                                    resource_server_name,
                                                                                    method, path, server_type)
            return isamAppliance.invoke_delete(
                "Delete an existing API Access Control Resource",
                url,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_selection(isamAppliance, instance_name, resource_server_name, resources, command="DELETE",
                     server_type="standard", check_mode=False, force=False):
    """
    Delete a selection of API Access Control Resources
    """
    found_any = False
    new_list_resources = []
    ret_obj = get_all(isamAppliance, instance_name, resource_server_name)
    warnings = ret_obj['warnings']

    for resource in resources:
        found = False
        for obj in ret_obj['data']:
            if obj['id'] == resource:
                found = True
                found_any = True
        if found is False:
            warnings.append("Did not find resource {0} to delete".format(resource))
        else:
            new_list_resources.append(resource)

    if force is True or found_any is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Delete a selection of API Access Control Resources",
                "{0}/{1}/server{2}/resource?server_type={3}".format(uri, instance_name, resource_server_name,
                                                                    server_type),
                {
                    'command': command,
                    'resources': new_list_resources
                },
                requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_all(isamAppliance, instance_name, resource_server_name, server_type='standard',
               check_mode=False, force=False):
    """
    Delete all existing API Access Control Resources
    """

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    resource_exist, warnings = _check_all_resource(isamAppliance, instance_name, resource_server_name)

    if force is True or resource_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete all existing API Access Control Resources",
                "{0}/{1}/server{2}/resource?server_type={3}".format(uri, instance_name, resource_server_name,
                                                                    server_type),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def export_all(isamAppliance, instance_name, resource_server_name, file_path, check_mode=False, force=False):
    """
    Exporting all existing API Access Control Resources
    """
    if os.path.exists(file_path) is True:
        warn_str = "File {0} already exists".format(file_path)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)
    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    if force is True or server_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            url = "{0}/{1}/server{2}/resource?export=true".format(uri, instance_name, resource_server_name)
            return isamAppliance.invoke_get_file(
                "Exporting all existing API Access Control Resources",
                url,
                file_path,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def export_file(isamAppliance, instance_name, resource_server_name, method, path, file_path,
                server_type="standard", check_mode=False, force=False):
    """
    Exporting an existing API Access Control Resource
    """
    if os.path.exists(file_path) is True:
        warn_str = "File {0} already exists".format(file_path)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)
    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    resource_exist, warnings = _check_resource_exist(isamAppliance, instance_name, resource_server_name, method, path)

    if force is True or resource_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if tools.version_compare(isamAppliance.facts["version"], "10.0.0") < 0:
                url = "{0}/{1}/server{2}/resource/{3}{4}?export=true&server_type={5}".format(uri, instance_name,
                                                                                             resource_server_name,
                                                                                             method, path, server_type)
            else:
                url = "{0}/{1}/server{2}/resource/{3}{2}{4}?export=true&server_type={5}".format(uri, instance_name,
                                                                                                resource_server_name,
                                                                                                method, path,
                                                                                                server_type)
            return isamAppliance.invoke_get_file(
                "Exporting an existing API Access Control Resource",
                url,
                file_path,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def import_file(isamAppliance, instance_name, resource_server_name, filename,
                server_type="standard", check_mode=False, force=False):
    """
    Importing an API Access Control Resource(s)
    """
    if os.path.exists(filename) is False:
        warn_str = "File {0} does not exists".format(filename)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)
    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    if force is True or server_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            url = "{0}/{1}/server{2}/resource?server_type={3}".format(uri, instance_name,
                                                                      resource_server_name, server_type)

            return isamAppliance.invoke_post_files(
                "Importing an API Access Control Resource(s)", url,
                [
                    {
                        'file_formfield': 'config_file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {}, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_instance_exist(isamAppliance, instance_name):
    ret_obj = ibmsecurity.isam.web.api_access_control.resources.instances.get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == instance_name:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check_server_exist(isamAppliance, instance_name, junction_point):
    ret_obj = ibmsecurity.isam.web.api_access_control.resources.servers.get_all(isamAppliance, instance_name)
    for obj in ret_obj['data']:
        if obj['name'] == junction_point:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check_resource_exist(isamAppliance, instance_name, resource_server_name, method, path):
    ret_obj = get_all(isamAppliance, instance_name, resource_server_name)
    warnings = ret_obj['warnings']
    if tools.version_compare(isamAppliance.facts["version"], "10.0.0") < 0:
        path_name = path
    else:
        path_name = "{0}{1}".format(resource_server_name, path)
    for obj in ret_obj['data']:
        if obj['method'] == method and obj['path'] == path_name:
            return True, warnings

    return False, warnings


def _check_all_resource(isamAppliance, instance_name, resource_server_name):
    ret_obj = get_all(isamAppliance, instance_name, resource_server_name)
    warnings = ret_obj['warnings']

    if ret_obj['data'] != []:
        return True, warnings
    else:
        return False, warnings


def _check_all_servers(isamAppliance, instance_name):
    ret_obj = ibmsecurity.isam.web.api_access_control.resources.servers.get_all(isamAppliance, instance_name)
    warnings = ret_obj['warnings']

    if ret_obj['data'] != []:
        return True, warnings
    else:
        return False, warnings


def _check_list_resource(isamAppliance, instance_name, resource_server_name, resources):
    ret_obj = get_all(isamAppliance, instance_name, resource_server_name)
    warnings = ret_obj['warnings']
    non_exist = False

    for resource in resources:
        found = False
        for obj in ret_obj['data']:
            if obj['id'] == resource:
                found = True
        if found is False:
            non_exist = True
            warnings.append("Did not find resource {0}".format(resource))

    if non_exist is False:
        return True, ret_obj['warnings']
    else:
        return False, warnings


def _check_list_servers(isamAppliance, instance_name, resource_servers):
    ret_obj = ibmsecurity.isam.web.api_access_control.resources.servers.get_all(isamAppliance, instance_name)
    warnings = ret_obj['warnings']
    non_exist = False

    for server in resource_servers:
        found = False
        for obj in ret_obj['data']:
            if obj['name'] == server:
                found = True
        if found is False:
            non_exist = True
            warnings.append("Did not find resource server {0}".format(server))

    if non_exist is False:
        return True, ret_obj['warnings']
    else:
        return False, warnings


def _check_resource_content(isamAppliance, instance_name, resource_server_name, method, path, policy,
                            cors_policy, static_response_headers, url_aliases, name, rate_limiting_policy,
                            documentation):
    ret_obj = get(isamAppliance, instance_name, resource_server_name, path, method)
    current_data = ret_obj['data']
    del current_data['id']

    json_data = {
        'method': method,
        'policy': policy,
        'cors_policy': cors_policy,
        'static_response_headers': static_response_headers,
        'url_aliases': url_aliases
    }

    if tools.version_compare(isamAppliance.facts["version"], "10.0.0") < 0:
        json_data['path'] = path
    else:
        json_data['path'] = "{0}{1}".format(resource_server_name, path)

    if name is not None:
        json_data['name'] = name

    if rate_limiting_policy is not None:
        json_data['rate_limiting_policy'] = rate_limiting_policy

    if documentation is not None:
        json_data['documentation'] = documentation

    sorted_obj1 = tools.json_sort(json_data)
    logger.debug("Sorted sorted_obj1: {0}".format(sorted_obj1))
    sorted_obj2 = tools.json_sort(current_data)
    logger.debug("Sorted sorted_obj2: {0}".format(sorted_obj2))

    if sorted_obj1 != sorted_obj2:
        logger.info("Changes detected, update needed.")
        return True, ret_obj['warnings'], json_data

    return False, ret_obj['warnings'], json_data


def compare_one_server(isamAppliance1, isamAppliance2, instance1_name, server1_name, instance2_name=None,
                       server2_name=None):
    """
    Compare resources between two appliances
    """

    if instance2_name is None or instance2_name == '':
        instance2_name = instance1_name

    if server2_name is None or server2_name == '':
        server2_name = server1_name

    resources1 = get_all(isamAppliance1, instance_name=instance1_name, resource_server_name=server1_name)
    resources2 = get_all(isamAppliance2, instance_name=instance2_name, resource_server_name=server2_name)

    return tools.json_compare(resources1, resources2)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare resources between two appliances
    """

    app1_instances = ibmsecurity.isam.web.api_access_control.resources.instances.get_all(isamAppliance1)
    app2_instances = ibmsecurity.isam.web.api_access_control.resources.instances.get_all(isamAppliance2)

    obj1 = []
    obj2 = []

    for inst1 in app1_instances['data']:
        servers = ibmsecurity.isam.web.api_access_control.resources.servers.get_all(isamAppliance1,
                                                                                    instance_name=inst1['name'])
        for srv in servers['data']:
            if "servers" in srv:
                srvlist = srv['servers'].split(";")
                new_str = None
                for value in srvlist:
                    if value.find("server_uuid") == -1 and \
                            value.find("current_requests") == -1 and \
                            value.find("total_requests") == -1:
                        if new_str is None:
                            new_str = value
                        else:
                            new_str = new_str + ";" + value
                srv['servers'] = new_str
            obj1.append(srv)
            resources1 = get_all(isamAppliance1, instance_name=inst1['name'], resource_server_name=srv['name'])
            obj1.extend(resources1['data'])

    for inst2 in app2_instances['data']:
        servers = ibmsecurity.isam.web.api_access_control.resources.servers.get_all(isamAppliance2,
                                                                                    instance_name=inst2['name'])
        for srv in servers['data']:
            if "servers" in srv:
                srvlist = srv['servers'].split(";")
                new_str = None
                for value in srvlist:
                    if value.find("server_uuid") == -1 and \
                            value.find("current_requests") == -1 and \
                            value.find("total_requests") == -1:
                        if new_str is None:
                            new_str = value
                        else:
                            new_str = new_str + ";" + value
                srv['servers'] = new_str
            obj2.append(srv)
            resources2 = get_all(isamAppliance2, instance_name=inst2['name'], resource_server_name=srv['name'])
            obj2.extend(resources2['data'])

    app1_instances['data'].extend(obj1)
    app2_instances['data'].extend(obj2)

    return tools.json_compare(app1_instances, app2_instances,
                              deleted_keys=['server_uuid', 'current_requests', 'total_requests'])



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.utilities import tools
import ibmsecurity.isam.web.reverse_proxy.junctions
from ibmsecurity.isam.web.reverse_proxy.junctions_config import server_fields
import json

logger = logging.getLogger(__name__)

uri = "/wga/reverseproxy"
requires_modules = ["wga"]
requires_version = None

def search(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port):
    ret_obj_new = isamAppliance.create_return_object()
    ret_obj = ibmsecurity.isam.web.reverse_proxy.junctions.get(isamAppliance, reverseproxy_id, junction_point)
    for s in ret_obj['data']['servers']:
        logger.debug("Servers in Junction server: {0} port: {1}".format(s['server_hostname'], s['server_port']))
        if str(server_hostname) == str(s['server_hostname']) and str(server_port) == str(s['server_port']):
            ret_obj_new['data'] = s['server_uuid']
            break

    return ret_obj_new

def get(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port):
    ret_obj_new = isamAppliance.create_return_object()
    ret_obj = ibmsecurity.isam.web.reverse_proxy.junctions.get(isamAppliance, reverseproxy_id, junction_point)
    for s in ret_obj['data']['servers']:
        logger.debug("Servers in Junction server: {0} port: {1}".format(s['server_hostname'], s['server_port']))
        if str(server_hostname) == str(s['server_hostname']) and str(server_port) == str(s['server_port']):
            ret_obj_new['data'] = s
            break
    return ret_obj_new

def add(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, junction_type="tcp", check_mode=False, force=False, warnings=[],
        **optionargs):
    """
    Adding a back-end server to an existing standard or virtual junctions

    :param isamAppliance:
    :param reverseproxy_id:
    :param junction_point:
    :param server_hostname:
    :param junction_type:
    :param server_port:
    :param virtual_hostname:
    :param virtual_https_hostname:
    :param server_dn:
    :param query_contents:
    :param stateful_junction:
    :param case_sensitive_url:
    :param case_insensitive_url:  #v1.0.6+
    :param windows_style_url:
    :param https_port:
    :param http_port:
    :param proxy_hostname:
    :param proxy_port:
    :param sms_environment:
    :param vhost_label:
    :param server_uuid:
    :param priority:
    :param server_cn:
    :param check_mode:
    :param force:
    :return:
    """
    # Search for the UUID of the junctioned server
    if not force:
        ret_obj = search(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port)

    if force or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            jct_srv_json = {
                "junction_point": junction_point,
                "junction_type": junction_type,
                "server_hostname": server_hostname,
                "server_port": server_port
            }
            for _k, _v in optionargs.items():
                if _v is not None:
                    jct_srv_json[_k] = _v
            if jct_srv_json.get('stateful_junction', None) is None:
                jct_srv_json['stateful_junction'] = 'no'
            if jct_srv_json.get('windows_style_url', None) is None:
                jct_srv_json['windows_style_url'] = 'no'

            # case_insensitive/case_sensitive
            if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") >= 0:
                # If no case_insensitive_url is passed, we take the old one and invert it.
                # Who thinks it's a good idea to make changes in an API like this ?
                if jct_srv_json.get('case_insensitive_url', None) is None:
                    case_sensitive_url = jct_srv_json.get('case_sensitive_url', None)
                    if case_sensitive_url is not None and case_sensitive_url.lower() == 'no':
                        jct_srv_json["case_insensitive_url"] = 'yes'
                    else:
                        jct_srv_json["case_insensitive_url"] = 'no' # default
                    jct_srv_json.pop("case_sensitive_url", None)
            else:
                jct_srv_json.pop("case_insensitive_url", None)

            if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                if jct_srv_json.get('server_cn', None) is not None:
                    warnings.append(
                        "Appliance at version: {0}, server_cn: {1} is not supported. Needs 10.0.2.0 or higher. Ignoring server_cn for this call.".format(
                            isamAppliance.facts["version"], server_cn))
                    jct_srv_json.pop("server_cn", None)

            if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                if jct_srv_json.get('priority', None) is not None:
                    warnings.append(
                        "Appliance at version: {0}, priority: {1} is not supported. Needs 10.0.2.0 or higher. Ignoring priority for this call.".format(
                            isamAppliance.facts["version"], priority))
                    jct_srv_json.pop("priority", None)
            else:
                if jct_srv_json.get('priority', None) is None:
                    warnings.append(
                        "Appliance at version: {0}, priority is required on 10.0.2 or higher".format(
                            isamAppliance.facts["version"]))
                    jct_srv_json['priority'] = "9"
            return isamAppliance.invoke_put(
                "Adding a back-end server to an existing standard or virtual junctions",
                "{0}/{1}/junctions".format(uri, reverseproxy_id), jct_srv_json)

    return isamAppliance.create_return_object()

def set(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, junction_type="tcp", check_mode=False, force=False, warnings=[],
        **optionargs):
    """
    Adding a back-end server to an existing standard or virtual junctions

    :param isamAppliance:
    :param reverseproxy_id:
    :param junction_point:
    :param server_hostname:
    :param junction_type:
    :param server_port:
    :param virtual_hostname:
    :param virtual_https_hostname:
    :param server_dn:
    :param query_contents:
    :param stateful_junction:
    :param case_sensitive_url:
    :param windows_style_url:
    :param https_port:
    :param http_port:
    :param proxy_hostname:
    :param proxy_port:
    :param sms_environment:
    :param vhost_label:
    :param server_uuid:
    :param server_cn:
    :param priority:
    :param check_mode:
    :param force:
    :return:
    """
    # load option args
    #   server_dn=None,
    #   stateful_junction='no', case_sensitive_url='no', windows_style_url='no', virtual_hostname=None,
    #   virtual_https_hostname=None, query_contents=None, https_port=None, http_port=None, proxy_hostname=None,
    #   proxy_port=None, sms_environment=None, vhost_label=None, server_uuid=None,
    #   server_cn=None, priority='9'

    # Search an existing server
    ret_obj = get(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port)
    exist_jct = {}
    jct_srv_json = {}
    if ret_obj:
        exist_jct = ret_obj.get('data', {})
        exist_jct.pop('current_requests', None)
        exist_jct.pop('total_requests', None)
        exist_jct.pop('operation_state', None)
        exist_jct.pop('server_state', None)
        exist_jct.pop('query_contents', None)

    for _k, _v in optionargs.items():
        # only keep valid arguments
        if _k in list(server_fields.keys()):
            jct_srv_json[_k] = _v
        else:
            logger.debug(f"Invalid input parameter used {_k}")
            warnings.append(f"Invalid input parameter used in function junctions_server.set() : {_k}")
    # add defaults
    #defaults_no = ["stateful_junction", "case_sensitive_url", "windows_style_url"]
    #for d in defaults_no:
    #    if jct_srv_json.get(d, None) is None:
    #        jct_srv_json[d] = 'no'
    # 10.0.0.2
    if jct_srv_json.get('priority', None) is not None:
        if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
            warnings.append(
                "Appliance at version: {0}, priority: {1} is not supported. Needs 10.0.2.0 or higher. Ignoring description for this call.".format(
                    isamAppliance.facts["version"], priority))
            jct_srv_json.pop('priority', None)
    else:
        if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") >= 0:
            warnings.append(
                "Appliance at version: {0}, priority is required".format(
                    isamAppliance.facts["version"]))
            jct_srv_json['priority'] = "9"

    # only compare values that are in the new request
    exist_jct = {k: v for k, v in exist_jct.items() if k in jct_srv_json.keys()}

    newJSON = json.dumps(jct_srv_json, skipkeys=True, sort_keys=True)
    logger.debug(f"\nSorted Desired  Junction {junction_point} - {server_hostname}:\n\n {newJSON}\n")

    oldJSON = json.dumps(exist_jct, skipkeys=True, sort_keys=True)
    logger.debug(f"\nSorted Current  Junction {junction_point} - {server_hostname}:\n\n {oldJSON}\n")

    jct_srv_json["junction_point"] = junction_point
    jct_srv_json["junction_type"] = junction_type
    jct_srv_json["server_hostname"] = server_hostname
    jct_srv_json["server_port"] = server_port

    if force or (newJSON != oldJSON):
        logger.debug(f"The JSONs are different.  We're going to add the servers.")
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Adding a back-end server to an existing standard or virtual junctions",
                "{0}/{1}/junctions".format(uri, reverseproxy_id), data=jct_srv_json, warnings=warnings)
    else:
        logger.debug("Servers are the same")
        return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, check_mode=False, force=False):
    """
    Deleting a standard or virtual junction's server

    :param isamAppliance:
    :param reverseproxy_id:
    :param junction_point:
    :param check_mode:
    :param force:
    :return:
    """
    # Search for the UUID of the junctioned server
    if force is False:
        ret_obj = search(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a standard or virtual junction's server",
                "{0}/{1}/junctions?junctions_id={2}&servers_id={3}".format(uri, reverseproxy_id, junction_point,
                                                                           ret_obj['data']))

    return isamAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isdsAppliance, check_mode=False, force=False):
    """
    Get current configured server type
    """
    return isdsAppliance.invoke_get("Retrieving Server Type", "/servertype_object")


def set(isdsAppliance, serverType="RDBM", check_mode=False, force=False):
    """
    Update Directory Server "server type"
    """
    if force is True or _check(isdsAppliance, serverType) is False:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_post(
                "Setting Server Type",
                "/servertype_object",
                {
                    "serverType": serverType
                })

    return isdsAppliance.create_return_object()


def _check(isdsAppliance, serverType):
    """
    Check if configured server type already set to specificed serverType
          Values           serverType
      ==================   ==========
      "Directory Server"     RDBM
      "Directory Proxy"      PROXY
      "Virtual Directory"    VD
    """
    ret_obj = get(isdsAppliance)

    if ret_obj['data']['DirectoryServerType'] == 'Directory Server':
        if serverType == 'RDBM':
            return True
        else:
            return False
    elif ret_obj['data']['DirectoryServerType'] == 'Directory Proxy':
        if serverType == 'PROXY':
            return True
        else:
            return False
    elif ret_obj['data']['DirectoryServerType'] == 'Virtual Directory':
        if serverType == 'VD':
            return True
        else:
            return False
    else:
        return True


def compare(isdsAppliance1, isdsAppliance2):
    """
    Compare server type values between two appliances
    """
    ret_obj1 = get(isdsAppliance1)
    ret_obj2 = get(isdsAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2)



import logging

logger = logging.getLogger(__name__)

# URI for this module
module_uri = "/silent_config"
requires_modules = None
requires_version = None
requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Getting the silent configuration flag
    """
    return isamAppliance.invoke_get("Getting the silent configuration flag", f"{module_uri}/flag",
                                    requires_version=requires_version, requires_modules=requires_modules,
                                    requires_model=requires_model)


def update(isamAppliance, flag, check_mode=False, force=False):
    """
    Setting the silent configuration flag
    """
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']
    if 'flag' in ret_obj['data']:
        if ret_obj['data']['flag'] != flag:
            check_value = True
        else:
            check_value = False
    else:
        check_value = False

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            json_data = {"flag": flag}
            return isamAppliance.invoke_put("Setting the silent configuration flag", f"{module_uri}/flag",
                                            json_data, requires_version=requires_version,
                                            requires_modules=requires_modules,
                                            requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _export(isamAppliance, uri, network_hostname, filename, network_1_1_ipv4_address, network_1_1_ipv4_netmask,
            network_1_1_ipv4_gateway, network_1_1_ipv6_address, network_1_1_ipv6_prefix, network_1_1_ipv6_gateway,
            include_policy, check_mode=False, force=False):
    """
    Generating a silent configuration
    """
    # create the request header for the post first
    headers = {"Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
               "Content-type": "application/x-www-form-urlencoded"
              }

    json_data = {
        "network.hostname": network_hostname,
        "network.1.1.ipv4.address": network_1_1_ipv4_address,
        "network.1.1.ipv4.netmask": network_1_1_ipv4_netmask,
        "network.1.1.ipv4.gateway": network_1_1_ipv4_gateway,
        "network.1.1.ipv6_address": network_1_1_ipv6_address,
        "network.1.1.ipv6.prefix": network_1_1_ipv6_prefix,
        "network.1.1.ipv6.gateway": network_1_1_ipv6_gateway,
        "include_policy": include_policy
    }

    post_data = ""
    for k, value in json_data.items():
        if value is not None:
            post_data = f"{post_data}{k}={value}&"

    # strip the last & added
    post_data = post_data[:-1]
    import os.path

    if force is True or os.path.exists(filename) is False:
        if check_mode is False:  # No point downloading a file if in check_mode

            ret_obj = isamAppliance.invoke_request("Generating a silent configuration", "post", uri=uri,
                                                   filename=filename, requires_modules=requires_modules,
                                                   requires_version=requires_version, data=post_data, headers=headers,
                                                   stream=True, requires_model=requires_model)
            # HTTP POST calls get flagged as changes - but no changes here
            if ret_obj['changed'] is True:
                ret_obj['changed'] = False

            return ret_obj

    return isamAppliance.create_return_object()


def export_iso(isamAppliance, network_hostname, filename=None, network_1_1_ipv4_address=None,
               network_1_1_ipv4_netmask=None, network_1_1_ipv4_gateway=None, network_1_1_ipv6_address=None,
               network_1_1_ipv6_prefix=None, network_1_1_ipv6_gateway=None, include_policy="false", check_mode=False,
               force=False):
    logger.info("Generating an ISO file for silent configuration.")
    uri = f"{module_uri}/create/iso"
    if filename is None:
        filename = f"{network_hostname}.iso"
        logger.debug(f"Using filename as: {filename}")

    return _export(isamAppliance, uri, network_hostname, filename, network_1_1_ipv4_address, network_1_1_ipv4_netmask,
                   network_1_1_ipv4_gateway, network_1_1_ipv6_address, network_1_1_ipv6_prefix,
                   network_1_1_ipv6_gateway, include_policy, check_mode, force)


def export_img(isamAppliance, network_hostname, filename=None, network_1_1_ipv4_address=None,
               network_1_1_ipv4_netmask=None, network_1_1_ipv4_gateway=None, network_1_1_ipv6_address=None,
               network_1_1_ipv6_prefix=None, network_1_1_ipv6_gateway=None, include_policy="false", check_mode=False,
               force=False):
    logger.info("Generating an IMG file for silent configuration.")

    uri = f"{module_uri}/create/usb"
    if filename is None:
        filename = f"{network_hostname}.img"
        logger.debug(f"Using filename as: {filename}")

    return _export(isamAppliance, uri, network_hostname, filename, network_1_1_ipv4_address, network_1_1_ipv4_netmask,
                   network_1_1_ipv4_gateway, network_1_1_ipv6_address, network_1_1_ipv6_prefix,
                   network_1_1_ipv6_gateway, include_policy, check_mode, force)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the LMI trace specification
    """
    return isamAppliance.invoke_get("Retrieving the LMI trace specification", "/admin_cfg/lmi_tracing")


def set(isamAppliance, trace_specification, check_mode=False, force=False):
    """
    Updating the LMI trace specification
    """
    current_obj = get(isamAppliance)
    current_specs = current_obj['data']['trace_specification']

    old_specs = current_specs.split(":")
    new_specs = trace_specification.split(":")

    old_specs.sort()
    new_specs.sort()

    if old_specs != new_specs or force is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Updating the LMI trace specification",
                "/admin_cfg/lmi_tracing", {'trace_specification': trace_specification})

    return isamAppliance.create_return_object()


def delete(isamAppliance, check_mode=False, force=False):
    """
    Resetting the LMI trace specification
    """
    current_obj = get(isamAppliance)
    current_specs = current_obj['data']['trace_specification']

    if force is True or current_specs != "*=config=enabled":
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Resetting the LMI trace specification",
                "/admin_cfg/lmi_tracing",
            )

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare lmi tracing between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get all snmp objects
    """
    return isamAppliance.invoke_get("Get all snmp objects",
                                    "/core/rsp_snmp_objs")


def get(isamAppliance, uuid, check_mode=False, force=False):
    """
    Get a specific snmp object
    """
    return isamAppliance.invoke_get("Get a specific snmp object",
                                    f"/core/rsp_snmp_objs/{uuid}")


def add(isamAppliance, name, trapAddress, trapCommunity, trapNotificationType=None, trapVersion='V1', trapPort=162,
        objType='snmp', username=None, authEnabled=None, authType=None, authPassPhrase=None, privEnabled=None,
        privType=None, privPassPhrase=None, informSnmpEngineID=None, informTimeout=None, comment='', check_mode=False,
        force=False):
    """
    Add a snmp object
    """
    if force is True or _check(isamAppliance, None, name, trapAddress, trapCommunity, trapNotificationType, trapVersion,
                               trapPort, objType, username, authEnabled, authType, authPassPhrase, privEnabled,
                               privType, privPassPhrase, informSnmpEngineID, informTimeout, comment) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Add a snmp object",
                "/core/rsp_snmp_objs/",
                {
                    'name': name,
                    'objType': objType,
                    'comment': comment,
                    'trapAddress': trapAddress,
                    'trapPort': trapPort,
                    'trapCommunity': trapCommunity,
                    'trapVersion': trapVersion,
                    'trapNotificationType': trapNotificationType,
                    'userName': username,
                    'authEnabled': authEnabled,
                    'authType': authType,
                    'authPassPhrase': authPassPhrase,
                    'privEnabled': privEnabled,
                    'privType': privType,
                    'privPassPhrase': privPassPhrase,
                    'informSnmpEngineID': informSnmpEngineID,
                    'informTimeout': informTimeout
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, uuid, name, trapAddress, trapCommunity, trapNotificationType=None, trapVersion='V1',
           trapPort=162, objType='snmp', username=None, authEnabled=None, authType=None, authPassPhrase=None,
           privEnabled=None, privType=None, privPassPhrase=None, informSnmpEngineID=None, informTimeout=None,
           comment='', check_mode=False, force=False):
    """
    Update a specific snmp object
    """
    if force is True or (
            _exists(isamAppliance, uuid) is True and _check(isamAppliance, uuid, name, trapAddress,
                                                            trapCommunity, trapNotificationType, trapVersion,
                                                            trapPort, objType, username, authEnabled, authType,
                                                            authPassPhrase, privEnabled, privType,
                                                            privPassPhrase, informSnmpEngineID, informTimeout,
                                                            comment) is False):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specific snmp object",
                f"/core/rsp_snmp_objs/{uuid}",
                {
                    'name': name,
                    'uuid': uuid,
                    'objType': objType,
                    'comment': comment,
                    'trapAddress': trapAddress,
                    'trapPort': trapPort,
                    'trapCommunity': trapCommunity,
                    'trapVersion': trapVersion,
                    'trapNotificationType': trapNotificationType,
                    'userName': username,
                    'authEnabled': authEnabled,
                    'authType': authType,
                    'authPassPhrase': authPassPhrase,
                    'privEnabled': privEnabled,
                    'privType': privType,
                    'privPassPhrase': privPassPhrase,
                    'informSnmpEngineID': informSnmpEngineID,
                    'informTimeout': informTimeout
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, uuid, check_mode=False, force=False):
    """
    Delete a snmp object
    """
    if force is True or _exists(isamAppliance, uuid) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a snmp object",
                f"/core/rsp_snmp_objs/{uuid}")

    return isamAppliance.create_return_object()


def _exists(isamAppliance, uuid):
    """
    Check if an uuid object exists

    :param isamAppliance:
    :param uuid:
    :return:
    """
    exists = False
    ret_obj = get_all(isamAppliance)

    for snmp in ret_obj['data']['snmpObjects']:
        if snmp['uuid'] == uuid:
            exists = True
            break

    return exists


def _check(isamAppliance, uuid, name, trapAddress, trapCommunity, trapNotificationType, trapVersion, trapPort, objType,
           username, authEnabled, authType, authPassPhrase, privEnabled, privType, privPassPhrase, informSnmpEngineID,
           informTimeout, comment):
    """
    Check if the snmp object exists and is the same - uuid=None means add versus delete

    NOTE: if UUID is not found that will be same as no match!!!
    """

    set_value = {
        'name': name,
        'uuid': uuid,
        'objType': objType,
        'comment': comment,
        'trapAddress': trapAddress,
        'trapPort': trapPort,
        'trapCommunity': trapCommunity,
        'trapVersion': trapVersion,
        'trapNotificationType': trapNotificationType,
        'userName': username,
        'authEnabled': authEnabled,
        'authType': authType,
        'authPassPhrase': authPassPhrase,
        'privEnabled': privEnabled,
        'privType': privType,
        'privPassPhrase': privPassPhrase,
        'informSnmpEngineID': informSnmpEngineID,
        'informTimeout': informTimeout
    }

    set_value = ibmsecurity.utilities.tools.json_sort(set_value)

    ret_obj = get_all(isamAppliance)
    for obj in ret_obj['data']['snmpObjects']:
        if uuid is None and obj['name'] == name:
            return True
        elif ibmsecurity.utilities.tools.json_sort(obj) == set_value:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare snmp objects between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']['snmpObjects']:
        del obj['uuid']
    for obj in ret_obj2['data']['snmpObjects']:
        del obj['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])



import ibmsecurity.utilities.tools
import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

module_uri = "/isam/felb/configuration/services"
requires_modules = None
requires_version = None
requires_model = "Appliance"


def get(isamAppliance, service_name, check_mode=False, force=False):
    """
    Retrieving Layer Configuration
    """
    return isamAppliance.invoke_get("Retrieving Layer Configuration", f"{module_uri}/{service_name}/layer",
                                    requires_version=requires_version, requires_modules=requires_modules,
                                    requires_model=requires_model)


def update(isamAppliance, service_name, type, layer7_secure, layer7_ssl_label, layer7_cookie, check_mode=False,
           force=False):
    """
    Updates specified service name layer
    """

    json_data = {'type': type}
    if type == "layer7":
        json_data['layer7_secure'] = layer7_secure
        json_data['layer7_cookie'] = layer7_cookie
        json_data['layer7_ssl_label'] = layer7_ssl_label

    change_required, warnings = _check(isamAppliance, service_name, json_data)

    if force is True or change_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating Service Layer", f"{module_uri}/{service_name}/layer",
                                            json_data, requires_version=requires_version,
                                            requires_modules=requires_modules, requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, service_name, json_data):
    """
    Checks update for idempotency
    """

    ret_obj = get(isamAppliance, service_name)
    warnings = ret_obj['warnings']

    if ret_obj['data'] == {}:
        return False, warnings

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(json_data)
    logger.debug(f"Sorted Existing Data:{sorted_ret_obj}")
    logger.debug(f"Sorted Desired  Data:{sorted_json_data}")

    if sorted_ret_obj != sorted_json_data:
        return True, warnings
    else:
        return False, warnings


def compare(isamAppliance1, service_name1, isamAppliance2, service_name2):
    """
    Compare cluster configuration between two appliances
    """
    ret_obj1 = get(isamAppliance1, service_name1)
    ret_obj2 = get(isamAppliance2, service_name2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)


def execute(isamAppliance, isamUser, commands, admin_domain='Default'):
    """
    Execute a pdadmin command
    """
    logger.debug('user is: ' + isamUser.username)
    if (isinstance(commands, basestring)):
        import ast
        commands = ast.literal_eval(commands)
    return isamAppliance.invoke_post("Execute pdadmin commands", "/isam/pdadmin/",
                                     {
                                         "admin_id": isamUser.username,
                                         "admin_pwd": isamUser.password,
                                         "commands": commands,
                                         "admin_domain": admin_domain
                                     })



import logging
import json
from ibmsecurity.utilities import tools
from io import open

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/policies"
uri_json = "/iam/access/v8/policies/json"  # new in 10.0.6.0


def get_all(
    isamAppliance,
    filter=None,
    sortBy=None,
    count=None,
    formatting="xml",
    check_mode=False,
    force=False,
):
    """
    Retrieve a list of policies

    formatting can be xml or json (actually, if it's not json, it's going to return xml) for version 10.0.6+
    """
    warnings = []
    if formatting == "json":
        if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") < 0:
            warnings.append(
                f"Appliance is at version: {isamAppliance.facts['version']}. JSON format not supported unless at least 10.0.6.0. Setting to xml."
            )
            formatting = "xml"
    if formatting == "json":
        return isamAppliance.invoke_get(
            "Retrieve a list of policies (JSON)",
            f"{uri_json}/{tools.create_query_string(filter=filter, sortBy=sortBy, count=count)}",
            warnings=warnings,
        )
    else:
        return isamAppliance.invoke_get(
            "Retrieve a list of policies",
            f"{uri}/{tools.create_query_string(filter=filter, sortBy=sortBy)}",
            warnings=warnings,
        )


def get(isamAppliance, name, formatting="xml", check_mode=False, force=False):
    """
    Retrieve a specific policy
    """
    ret_obj = search(
        isamAppliance,
        name=name,
        formatting=formatting,
        check_mode=check_mode,
        force=force,
    )
    pol_id = ret_obj["data"]

    if pol_id == {}:
        logger.info(f"Policy {name} had no match, skipping retrieval.")
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, pol_id, formatting=formatting)


def _get(isamAppliance, pol_id, formatting="xml"):
    """
    Retrieve a specific access control policy by id
    """
    warnings = []
    if formatting == "json":
        if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") < 0:
            warnings.append(
                "Appliance is at version: {0}. JSON format not supported unless at least 10.0.6.0. Setting to xml.".format(                    isamAppliance.facts["version"]
                )
            )
            formatting = "xml"
    if formatting == "json":
        return isamAppliance.invoke_get(
            "Retrieve a specific policy (JSON)",
            f"{uri_json}/{pol_id}",
            warnings=warnings,
        )
    else:
        return isamAppliance.invoke_get(
            "Retrieve a specific policy",
            f"{uri}/{pol_id}",
            warnings=warnings,
        )


def export_xacml(
    isamAppliance,
    name,
    filename,
    formatting="xml",
    overwrite=False,
    check_mode=False,
    force=False,
):
    """
    Export XACML for a specific policy
    """
    warnings = []
    import os.path

    if formatting == "json":
        if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") < 0:
            warnings.append(
                f"Appliance is at version: {isamAppliance.facts['version']}. JSON format not supported unless at 10.0.6.0 or higher. Setting to xml."
            )
            formatting = "xml"
    if not force:
        ret_obj = get(
            isamAppliance,
            name=name,
            formatting=formatting,
            check_mode=check_mode,
            force=force,
        )

    if force or (ret_obj["data"] != {} and not os.path.exists(filename)):
        logger.debug("\n\nDOWNLOADING\n\n")
        if not check_mode:  # No point downloading a file if in check_mode
            f = open(filename, "w")
            if formatting == "json":
                logger.debug(f"JSON DATA:\n{ret_obj['data']['policy']}")
                f.write(json.dumps(ret_obj["data"]["policy"]))
            else:
                f.write(ret_obj["data"]["policy"])
    elif ret_obj["data"] != {} and os.path.exists(filename):
        warnings.append(f"File {filename} exists already. Overwriting = {overwrite}")
        if overwrite:
            if formatting == "json":
                logger.debug(f"JSON DATA:\n{ret_obj['data']['policy']}")
                new_policy = json.dumps(ret_obj["data"]["policy"])
            else:
                new_policy = ret_obj["data"]["policy"]
            # Compare checksums
            with open(filename, "r") as myfile:
                policy = myfile.read()

            if policy == new_policy:
                warnings.append("The content is the same.")
            else:
                myfile.close()
                f = open(filename, "w")
                f.write(new_policy)
    return isamAppliance.create_return_object(warnings=warnings)


def search(isamAppliance, name, formatting="xml", force=False, check_mode=False):
    """
    Search policy id by name
    """
    ret_obj = get_all(isamAppliance, formatting=formatting)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj["data"]:
        if obj["name"] == name:
            logger.info(f"Found Policy {name} id: {obj['id']}")
            return_obj["data"] = obj["id"]
            return_obj["rc"] = 0

    return return_obj


def set_file(
    isamAppliance,
    name,
    attributesrequired,
    policy_file,
    description="",
    dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os",
    predefined=False,
    new_name=None,
    formatting="xml",
    check_mode=False,
    force=False,
):
    import re

    # Read policy from file and call set()
    # The policy file no longer needs to contain the xml as a single line.

    with open(policy_file, "r") as myfile:
        policy = myfile.read()
        if formatting == "xml":
            # // 4. remove \n before an end tag
            policy = re.sub(r"([\w0-9>*$-._]+)\n\s+", r"\1", policy)
            # // 1. remove all white space preceding a begin element tag:
            policy = re.sub(r"[\n\s]+(<[^/])", r"\1", policy)
            # // 2. remove all white space following an end element tag:
            policy = re.sub(r"(</[a-zA-Z0-9-_.:]+>)\s+", r"\1", policy)
            # // 3. remove all white space following an empty element tag
            policy = re.sub(r"(/>)\s+", r"\1", policy)
            # // 5. remove remaining /n+any whitespace
            policy = re.sub(r"\n\s+", " ", policy)
        else:
            policy = json.loads(policy)

    return set(
        isamAppliance,
        name,
        attributesrequired,
        policy=policy,
        description=description,
        dialect=dialect,
        predefined=predefined,
        new_name=new_name,
        formatting=formatting,
        check_mode=check_mode,
        force=force,
    )


def set(
    isamAppliance,
    name,
    attributesrequired,
    policy,
    description="",
    dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os",
    predefined=False,
    new_name=None,
    formatting="xml",
    check_mode=False,
    force=False,
):
    """
    Creating or Modifying a Policy
    """
    if (search(isamAppliance, name=name, formatting=formatting))["data"] == {}:
        # Force the add - we already know policy does not exist
        logger.info(f"Policy {name} had no match, requesting to add new one.")
        return add(
            isamAppliance,
            name,
            attributesrequired,
            policy,
            description=description,
            dialect=dialect,
            predefined=predefined,
            formatting=formatting,
            check_mode=check_mode,
            force=True,
        )
    else:
        # Update request
        logger.info(f"Policy {name} exists, requesting to update.")
        return update(
            isamAppliance,
            name,
            attributesrequired,
            policy,
            description=description,
            dialect=dialect,
            predefined=predefined,
            new_name=new_name,
            formatting=formatting,
            check_mode=check_mode,
            force=force,
        )


def add(
    isamAppliance,
    name,
    attributesrequired,
    policy,
    description="",
    dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os",
    predefined=False,
    formatting="xml",
    check_mode=False,
    force=False,
):
    """
    Create a new Policy is not supported through the API in xml format

    The documentation basically states that you can only create a new policy by `duplicating` an existing policy (which means the output from a GET command)
    """
    if not force:
        ret_obj = search(isamAppliance, name, formatting=formatting)

    if force or ret_obj["data"] == {}:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            warnings = []
            if formatting == "json":
                if (
                    tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") < 0
                ):
                    warnings.append(
                        f"Appliance is at version: {isamAppliance.facts['version']}. JSON format not supported unless at least 10.0.6.0. Setting to xml."
                    )
                    formatting = "xml"
            if formatting == "json":
                logger.debug("\n\nJSON\n\n")
                json_data = {
                    "name": name,
                    "attributesRequired": attributesrequired,
                    "description": description,
                    "predefined": predefined,
                    "policy": policy,
                    "dialect": dialect,
                }
                return isamAppliance.invoke_post(
                    "Create a new Policy (JSON)", uri_json, json_data, warnings=warnings
                )
            else:
                # json.loads fails all the time with little information, also when using the output of GET (as supported/documented)
                #   to make this work, you need to ' double escape ' the xml in the output of the GET (so it looks like this : `<Policy PolicyId=\\"urn:ibm:security:rule-container:4\\"` )
                #   alternatively, use the new full json API of course, or use an xml file using set_file
                #
                #       json_data = json.loads(policy)
                # So ... lets just to string based checks on the object.
                # For it to be valid XML, it has to start with <?xml (this is not true, but it is what the output of ISVA looks like)
                if policy.startswith("<?xml"):
                    # this is xml
                    logger.info(f"Policy {name} only contains policy data")
                    json_data = {
                        "name": name,
                        "attributesrequired": attributesrequired,
                        "description": description,
                        "predefined": predefined,
                        "policy": policy,
                        "dialect": dialect,
                    }
                else:
                    # strip off the surrounding [ and ]
                    json_data = policy[1:-1]
                    logger.debug(f"\njson_data: {json_data}")
                    json_data = json.loads(json_data)
                    logger.info(f"Policy {name} contains full policy export")

                return isamAppliance.invoke_post(
                    "Create a new Policy", uri, json_data, warnings=warnings
                )

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a Policy
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    mech_id = ret_obj["data"]

    if mech_id == {}:
        logger.info(f"Policy {name} not found, skipping delete.")
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a Policy", f"{uri}/{mech_id}"
            )

    return isamAppliance.create_return_object()


def update(
    isamAppliance,
    name,
    attributesrequired,
    policy,
    description=None,
    dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os",
    predefined=False,
    new_name=None,
    formatting="xml",
    check_mode=False,
    force=False
):
    """
    Update a specified policy
    """
    warnings = []
    if formatting == "json":
        if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") < 0:
            warnings.append(
                f"Appliance is at version: {isamAppliance.facts['version']}. JSON format not supported unless at least 10.0.6.0. Setting to xml."
            )
            formatting = "xml"

    pol_id, update_required, json_data = _check(
        isamAppliance,
        name,
        attributesrequired,
        policy,
        description=description,
        dialect=dialect,
        predefined=predefined,
        new_name=new_name,
        formatting=formatting,
        warnings=warnings
    )

    if pol_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError

        raise IBMError("999", f"Cannot update data for unknown policy: {name}")

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if formatting == "json":
                return isamAppliance.invoke_put(
                    "Update a specified policy (JSON)",
                    f"{uri_json}/{pol_id}",
                    json_data,
                )
            else:
                return isamAppliance.invoke_put(
                    "Update a specified policy",
                    f"{uri}/{pol_id}",
                    json_data,
                )

    return isamAppliance.create_return_object()


def _check(
    isamAppliance,
    name,
    attributesrequired,
    policy,
    description=None,
    dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os",
    predefined=False,
    new_name=None,
    formatting="xml",
    warnings=[]
):
    """
    Check and return True if update needed
    """
    update_required = False
    if formatting == "json":
        if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") < 0:
            warnings.append(
                f"Appliance is at version: {isamAppliance.facts['version']}. JSON format not supported unless at least 10.0.6.0. Setting to xml."
            )
            formatting = "xml"

    if formatting == "json":
        logger.info(f"Loading JSON formatted Policy with name {name}")
        json_data = {
            "attributesRequired": attributesrequired,
            "policy": policy,
            "dialect": dialect,
            "predefined": predefined,
        }
    else:
        try:
            json_data = json.loads(policy)[0]
            logger.info(f"Policy {name} contains full policy export")
            logger.info(json_data)
            json_data.pop("id", None)
            json_data.pop("datecreated", None)
            json_data.pop("dateCreated", None)
            json_data.pop("lastmodified", None)
            json_data.pop("lastModified", None)
            json_data.pop("userlastmodified", None)
            json_data.pop("userLastModified", None)
            json_data.pop("predefined", None)
        except json.decoder.JSONDecodeError:
            logger.info(f"Policy {name} only contains policy data")
            json_data = {
                "attributesrequired": attributesrequired,
                "policy": policy,
                "dialect": dialect,
                "predefined": predefined,
            }

    ret_obj = get(isamAppliance, name, formatting=formatting)
    if ret_obj["data"] == {}:
        logger.warning("Policy not found, returning no update required.")
        return None, update_required, json_data
    else:
        pol_id = ret_obj["data"]["id"]
        if new_name is not None:
            json_data["name"] = new_name
        else:
            json_data["name"] = name
        if description is not None:
            json_data["description"] = description
        else:
            ret_obj["data"].pop("description", None)
        ret_obj["data"].pop("id", None)
        ret_obj["data"].pop("datecreated", None)
        ret_obj["data"].pop("dateCreated", None)
        ret_obj["data"].pop("lastmodified", None)
        ret_obj["data"].pop("lastModified", None)
        ret_obj["data"].pop("userlastmodified", None)
        ret_obj["data"].pop("userLastModified", None)

        sorted_json_data = tools.json_sort(json_data)
        logger.debug(f"\n\nSorted input: {sorted_json_data}")
        sorted_ret_obj = tools.json_sort(ret_obj["data"])
        logger.debug(f"\n\nSorted existing data: {sorted_ret_obj}")
        if sorted_ret_obj != sorted_json_data:
            logger.info("\n\nChanges detected, update needed.\n\n")
            update_required = True

        return pol_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2, formatting="json"):
    """
    Compare Policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1, formatting=formatting)
    ret_obj2 = get_all(isamAppliance2, formatting=formatting)

    for obj in ret_obj1["data"]:
        obj.pop("id")
        obj.pop("datecreated", None)
        obj.pop("dateCreated", None)
        obj.pop("lastmodified", None)
        obj.pop("lastModified", None)
        obj.pop("userlastmodified", None)
        obj.pop("userLastModified", None)
        ret_obj = get(isamAppliance1, obj["name"], formatting=formatting)
        obj["policy"] = ret_obj["data"]["policy"]
    for obj in ret_obj2["data"]:
        obj.pop("id")
        obj.pop("datecreated", None)
        obj.pop("dateCreated", None)
        obj.pop("lastmodified", None)
        obj.pop("lastModified", None)
        obj.pop("userlastmodified", None)
        obj.pop("userLastModified", None)
        ret_obj = get(isamAppliance2, obj["name"], formatting=formatting)
        obj["policy"] = ret_obj["data"]["policy"]

    return tools.json_compare(
        ret_obj1,
        ret_obj2,
        deleted_keys=["id", "userlastmodified", "lastmodified", "datecreated"],
    )



import logging

logger = logging.getLogger(__name__)


def get(isamAppliance, file_path='/', recursive=True, check_mode=False, force=False):
    """
    Retrieving the contents of a directory from the file downloads area
    """
    return isamAppliance.invoke_get("Retrieving the contents of a directory from the file downloads area",
                                    f"/isam/downloads/{file_path}/?recursive={recursive}")


def export_file(isamAppliance, file_path, filename, check_mode=False, force=False):
    """
    Downloading a file from the file downloads area
    """
    import os.path

    if force is True or os.path.exists(filename) is False:
        if check_mode is False:  # No point downloading a file if in check_mode
            # NOTE: this call requires no headers
            return isamAppliance.invoke_get_file(
                "Downloading a file from the file downloads area",
                f"/isam/downloads/{file_path}?type=File&browser=&",
                filename, True)

    return isamAppliance.create_return_object()



import logging

from ibmsecurity.isam.fed import attribute_source
from ibmsecurity.utilities import tools
from ibmsecurity.isam.aac import access_policy

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/definitions"
requires_modules = ["mga", "federation"]
requires_version = None

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of API protection definitions
    """
    return isamAppliance.invoke_get("Retrieve a list of API protection definitions", uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific API protection definition
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    defn_id = ret_obj['data']
    warnings = ret_obj["warnings"]

    if defn_id == {}:
        logger.info(f"Definition {name} had no match, skipping retrieval.")
        warnings.append(f"Definition Name {name} had no match.")
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        return _get(isamAppliance, defn_id)


def _get(isamAppliance, defn_id):
    return isamAppliance.invoke_get("Retrieve a specific API protection definition",
                                    f"{uri}/{defn_id}", requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Search definition id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info(f"Found definition {name} id: {obj['id']}")
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def add(isamAppliance, name, description="", accessPolicyName=None, grantTypes=["AUTHORIZATION_CODE"],
        tcmBehavior="NEVER_PROMPT",
        accessTokenLifetime=3600, accessTokenLength=20, enforceSingleUseAuthorizationGrant=False,
        authorizationCodeLifetime=300, authorizationCodeLength=30, issueRefreshToken=True, refreshTokenLength=40,
        maxAuthorizationGrantLifetime=604800, enforceSingleAccessTokenPerGrant=False,
        enableMultipleRefreshTokensForFaultTolerance=False, pinPolicyEnabled=False, pinLength=4,
        tokenCharSet="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", oidc=None, check_mode=False,
        force=False):
    """
    Create an API protection definition
    """
    if (isinstance(grantTypes, basestring)):
        import ast
        grantTypes = ast.literal_eval(grantTypes)

    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    warnings = ret_obj["warnings"]

    if force is True or ret_obj["data"] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            json_data = {
                "name": name,
                "description": description,
                "grantTypes": grantTypes,
                "tcmBehavior": tcmBehavior,
                "accessTokenLifetime": int(accessTokenLifetime),
                "accessTokenLength": int(accessTokenLength),
                "enforceSingleUseAuthorizationGrant": enforceSingleUseAuthorizationGrant,
                "authorizationCodeLifetime": int(authorizationCodeLifetime),
                "authorizationCodeLength": int(authorizationCodeLength),
                "issueRefreshToken": issueRefreshToken,
                "refreshTokenLength": int(refreshTokenLength),
                "maxAuthorizationGrantLifetime": int(maxAuthorizationGrantLifetime),
                "enforceSingleAccessTokenPerGrant": enforceSingleAccessTokenPerGrant,
                "enableMultipleRefreshTokensForFaultTolerance": enableMultipleRefreshTokensForFaultTolerance,
                "pinPolicyEnabled": pinPolicyEnabled,
                "pinLength": int(pinLength),
                "tokenCharSet": tokenCharSet
            }
            if accessPolicyName is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        f"Appliance at version: {isamAppliance.facts['version']}, access policy: {oidc} is not supported. Needs 9.0.4.0 or higher. Ignoring access policy for this call.")
                    accessPolicyName = None
                else:
                    ap_ret_obj = access_policy.search(isamAppliance, accessPolicyName, check_mode=check_mode, force=force)
                    if ap_ret_obj['data'] == {}:
                        warnings = ap_ret_obj["warnings"]
                        warnings.append(
                            f"Access Policy {accessPolicyName} is not found. Cannot add definition.")
                        return isamAppliance.create_return_object(warnings=warnings)
                    else:
                        json_data["accessPolicyId"] = int(ap_ret_obj['data'])

            if oidc is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        f"Appliance at version: {isamAppliance.facts['version']}, oidc: {oidc} is not supported. Needs 9.0.4.0 or higher. Ignoring oidc for this call.")
                else:
                    if 'attributeSources' in oidc:
                        oidc['attributeSources'] = _map_oidc_attributeSources(isamAppliance, oidc['attributeSources'], check_mode, force)
                    json_data["oidc"] = oidc
                if 'dynamicClients' in json_data['oidc']:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
                        warnings.append(
                            f"Appliance at version: {isamAppliance.facts['version']}, dynamicClients: {json_data['oidc']['dynamicClients']} is not supported. Needs 9.0.5.0 or higher. Ignoring dynamicClients for this call.")
                        del json_data['oidc']['dynamicClients']
                if 'issueSecret' in json_data['oidc']:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
                        warnings.append(
                            f"Appliance at version: {isamAppliance.facts['version']}, issueSecret: {json_data['oidc']['issueSecret']} is not supported. Needs 9.0.5.0 or higher. Ignoring issueSecret for this call.")
                        json_data['oidc'].pop('issueSecret', None)
                if 'includeIssInAuthResp' in json_data['oidc']:
                    if tools.version_compare(isamAppliance.facts["version"], "10.0.8.0") < 0:
                        warnings.append(
                            f"Appliance at version: {isamAppliance.facts['version']}, issueSecret: {json_data['oidc']['includeIssInAuthResp']} is not supported. Needs 10.0.8.0 or higher. Ignoring includeIssInAuthResp for this call.")
                        json_data['oidc'].pop('includeIssInAuthResp', None)

            return isamAppliance.invoke_post(
                "Create an API protection definition", uri,
                json_data, requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an API protection definition
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    defn_id = ret_obj['data']
    warnings = ret_obj["warnings"]

    if defn_id == {}:
        logger.info(f"Definition {name} not found, skipping delete.")
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Delete an API protection definition",
                f"{uri}/{defn_id}", requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, name, description="", accessPolicyName=None, grantTypes=["AUTHORIZATION_CODE"],
           tcmBehavior="NEVER_PROMPT",
           accessTokenLifetime=3600, accessTokenLength=20, enforceSingleUseAuthorizationGrant=False,
           authorizationCodeLifetime=300, authorizationCodeLength=30, issueRefreshToken=True, refreshTokenLength=40,
           maxAuthorizationGrantLifetime=604800, enforceSingleAccessTokenPerGrant=False,
           enableMultipleRefreshTokensForFaultTolerance=False, pinPolicyEnabled=False, pinLength=4,
           tokenCharSet="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", oidc=None, check_mode=False,
           force=False):
    """
    Update a specified API protection definition
    """
    ret_obj = get(isamAppliance, name)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append(f"Definiton {name} not found, skipping update.")
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        defn_id = ret_obj["data"]["id"]

    needs_update = False
    json_data = {
        "name": name,
        "description": description,
        "grantTypes": grantTypes,
        "tcmBehavior": tcmBehavior,
        "accessTokenLifetime": int(accessTokenLifetime),
        "accessTokenLength": int(accessTokenLength),
        "enforceSingleUseAuthorizationGrant": enforceSingleUseAuthorizationGrant,
        "authorizationCodeLifetime": int(authorizationCodeLifetime),
        "authorizationCodeLength": int(authorizationCodeLength),
        "issueRefreshToken": issueRefreshToken,
        "refreshTokenLength": int(refreshTokenLength),
        "maxAuthorizationGrantLifetime": int(maxAuthorizationGrantLifetime),
        "enforceSingleAccessTokenPerGrant": enforceSingleAccessTokenPerGrant,
        "enableMultipleRefreshTokensForFaultTolerance": enableMultipleRefreshTokensForFaultTolerance,
        "pinPolicyEnabled": pinPolicyEnabled,
        "pinLength": int(pinLength),
        "tokenCharSet": tokenCharSet
    }
    if accessPolicyName is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                f"Appliance at version: {isamAppliance.facts['version']}, access policy: {oidc} is not supported. Needs 9.0.4.0 or higher. Ignoring access policy for this call.")
            accessPolicyName = None
        else:
            ap_ret_obj = access_policy.search(isamAppliance, accessPolicyName, check_mode=check_mode, force=force)
            if ap_ret_obj['data'] == {}:
                warnings = ap_ret_obj["warnings"]
                warnings.append(
                    f"Access Policy {accessPolicyName} is not found. Cannot update definition.")
                return isamAppliance.create_return_object(warnings=warnings)
            else:
                json_data["accessPolicyId"] = int(ap_ret_obj['data'])

    if oidc is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                f"Appliance at version: {isamAppliance.facts['version']}, oidc: {oidc} is not supported. Needs 9.0.4.0 or higher. Ignoring oidc for this call.")
            oidc = None
        else:
            if 'attributeSources' in oidc:
                oidc['attributeSources'] = _map_oidc_attributeSources(isamAppliance, oidc['attributeSources'], check_mode, force)
            json_data["oidc"] = oidc

    if not force:

        ret_obj['data'].pop('datecreated', None)
        ret_obj['data'].pop('id', None)
        ret_obj['data'].pop('lastmodified', None)
        ret_obj['data'].pop('mappingRules', None)

        # Inspecting oidcConfig and remove missing or None attributes in returned object
        if oidc is not None and 'oidc' in ret_obj['data']:
            if 'enabled' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['enabled'] is None:
                del ret_obj['data']['oidc']['enabled']
            if 'iss' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['iss'] is None:
                del ret_obj['data']['oidc']['iss']
            if 'poc' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['poc'] is None:
                del ret_obj['data']['oidc']['poc']
            if 'lifetime' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['lifetime'] is None:
                del ret_obj['data']['oidc']['lifetime']
            if 'alg' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['alg'] is None:
                del ret_obj['data']['oidc']['alg']
            if 'db' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['db'] is None:
                del ret_obj['data']['oidc']['db']
            if 'cert' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['cert'] is None:
                del ret_obj['data']['oidc']['cert']
            if 'attributeSources' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['attributeSources'] is None:
                del ret_obj['data']['oidc']['attributeSources']

            # Inspecting oidcEncConfig and remove missing or None attributes in returned object
            if 'enc' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['enc'] is not None:
                if 'enabled' in ret_obj['data']['oidc']['enc'] and ret_obj['data']['oidc']['enc']['enabled'] is None:
                    del ret_obj['data']['oidc']['enc']['enabled']
                if 'alg' in ret_obj['data']['oidc']['enc'] and ret_obj['data']['oidc']['enc']['alg'] is None:
                    del ret_obj['data']['oidc']['enc']['alg']
                if 'enc' in ret_obj['data']['oidc']['enc'] and ret_obj['data']['oidc']['enc']['enc'] is None:
                    del ret_obj['data']['oidc']['enc']['enc']

            # For dynamicClients & issueSecret parameters
            #
            # If the values for dynamicClients or issueSecret are missing, then they are
            # considered to be of the value "false" by the appliance, this allows for old
            # configuration to be forward compatible, without the function of the
            # definition being changed by the same payload.
            if 'dynamicClients' in json_data['oidc']:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
                    warnings.append(
                        f"Appliance at version: {isamAppliance.facts['version']}, dynamicClients: {json_data['oidc']['dynamicClients']} is not supported. Needs 9.0.5.0 or higher. Ignoring dynamicClients for this call.")
                    del json_data['oidc']['dynamicClients']
            else:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") >= 0:
                    if 'dynamicClients' in ret_obj['data']['oidc'] and ret_obj['data']['oidc'][
                        'dynamicClients'] is False:
                        del ret_obj['data']['oidc']['dynamicClients']

            if 'issueSecret' in json_data['oidc']:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
                    warnings.append(
                        f"Appliance at version: {isamAppliance.facts['version']}, issueSecret: {json_data['oidc']['issueSecret']} is not supported. Needs 9.0.5.0 or higher. Ignoring issueSecret for this call.")
                    del json_data['oidc']['issueSecret']
            else:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") >= 0:
                    if 'issueSecret' in ret_obj['data']['oidc'] and ret_obj['data']['oidc']['issueSecret'] is False:
                        del ret_obj['data']['oidc']['issueSecret']

            if 'fapiCompliant' in ret_obj['data']['oidc'] and 'fapiCompliant' not in json_data['oidc']:
                del ret_obj['data']['oidc']['fapiCompliant']
            if 'oidcCompliant' in ret_obj['data']['oidc'] and 'oidcCompliant' not in json_data['oidc']:
                del ret_obj['data']['oidc']['oidcCompliant']

        if oidc is None and 'oidc' in ret_obj['data']:
            del ret_obj['data']['oidc']

        sorted_ret_obj = tools.json_sort(ret_obj['data'])
        sorted_json_data = tools.json_sort(json_data)
        logger.debug(f"Sorted Existing Data:{sorted_ret_obj}")
        logger.debug(f"Sorted Desired  Data:{sorted_json_data}")
        if sorted_ret_obj != sorted_json_data:
            needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update a specified API protection definition",
                f"{uri}/{defn_id}", json_data, requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, name, description="", accessPolicyName=None, grantTypes=["AUTHORIZATION_CODE"],
        tcmBehavior="NEVER_PROMPT",
        accessTokenLifetime=3600, accessTokenLength=20, enforceSingleUseAuthorizationGrant=False,
        authorizationCodeLifetime=300, authorizationCodeLength=30, issueRefreshToken=True, refreshTokenLength=40,
        maxAuthorizationGrantLifetime=604800, enforceSingleAccessTokenPerGrant=False,
        enableMultipleRefreshTokensForFaultTolerance=False, pinPolicyEnabled=False, pinLength=4,
        tokenCharSet="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", oidc=None, check_mode=False,
        force=False):
    """
    Creating or Modifying an API Protection Definition
    """
    ### Fix for issue #252 ###
    if oidc is not None and 'lifetime' in oidc and oidc['lifetime'] is not None:
        oidc['lifetime'] = int(oidc['lifetime'])

    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know policy does not exist
        logger.info(f"Definition {name} had no match, requesting to add new one.")
        return add(isamAppliance=isamAppliance, name=name, description=description, accessPolicyName=accessPolicyName,
                   grantTypes=grantTypes, tcmBehavior=tcmBehavior,
                   accessTokenLifetime=accessTokenLifetime, accessTokenLength=accessTokenLength,
                   enforceSingleUseAuthorizationGrant=enforceSingleUseAuthorizationGrant,
                   authorizationCodeLifetime=authorizationCodeLifetime, authorizationCodeLength=authorizationCodeLength,
                   issueRefreshToken=issueRefreshToken, refreshTokenLength=refreshTokenLength,
                   maxAuthorizationGrantLifetime=maxAuthorizationGrantLifetime,
                   enforceSingleAccessTokenPerGrant=enforceSingleAccessTokenPerGrant,
                   enableMultipleRefreshTokensForFaultTolerance=enableMultipleRefreshTokensForFaultTolerance,
                   pinPolicyEnabled=pinPolicyEnabled, pinLength=pinLength,
                   tokenCharSet=tokenCharSet, oidc=oidc, check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info(f"Definition {name} exists, requesting to update.")
        return update(isamAppliance=isamAppliance, name=name, description=description,
                      accessPolicyName=accessPolicyName,
                      grantTypes=grantTypes, tcmBehavior=tcmBehavior,
                      accessTokenLifetime=accessTokenLifetime, accessTokenLength=accessTokenLength,
                      enforceSingleUseAuthorizationGrant=enforceSingleUseAuthorizationGrant,
                      authorizationCodeLifetime=authorizationCodeLifetime,
                      authorizationCodeLength=authorizationCodeLength, issueRefreshToken=issueRefreshToken,
                      refreshTokenLength=refreshTokenLength,
                      maxAuthorizationGrantLifetime=maxAuthorizationGrantLifetime,
                      enforceSingleAccessTokenPerGrant=enforceSingleAccessTokenPerGrant,
                      enableMultipleRefreshTokensForFaultTolerance=enableMultipleRefreshTokensForFaultTolerance,
                      pinPolicyEnabled=pinPolicyEnabled, pinLength=pinLength,
                      tokenCharSet=tokenCharSet, oidc=oidc, check_mode=check_mode, force=force)


def _map_oidc_attributeSources(isamAppliance, attributeSources, check_mode=False, force=False):
    """
     Maps OIDC definition attributeSources from a string name to an ID
     :param isamAppliance: ISAM Appliance to act on
     :param attributeSources: Attribute sources to convert from string to integer
     :param check_mode: Ignored
     :param force: Ignored
     :return: attributeSources with attribute id mapped from attribute name
    """

    if attributeSources is not None:
        attr_lookup = None
        for source in attributeSources:
            # Only perform mapping if attributeSourceId is not an integer (this provides backwards compatibility)
            if not source['attributeSourceId'].isdigit():
                if not attr_lookup:
                    attr_lookup = attribute_source.get_all(isamAppliance, check_mode=check_mode, force=force)['data']
                for attr in attr_lookup:
                    if attr['name'] == source['attributeSourceId']:
                        source['attributeSourceId'] = attr['id']

    return attributeSources


def compare(isamAppliance1, isamAppliance2):
    """
    Compare API Protection Definitions between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
        del obj['datecreated']
        del obj['lastmodified']
        for rules in obj['mappingRules']:
            del rules['id']
    for obj in ret_obj2['data']:
        del obj['id']
        del obj['datecreated']
        del obj['lastmodified']
        for rules in obj['mappingRules']:
            del rules['id']

    return tools.json_compare(ret_obj1, ret_obj2,
                              deleted_keys=['id', 'datecreated', 'lastmodified',
                                            'mappingRules/id'])



import logging

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current log file language
    """
    return isamAppliance.invoke_get("Retrieve the current log file language",
                                    "/isam/applang/v1"
                                   )


def update(isamAppliance, id, check_mode=False, force=False):
    """
    Change the log file language
    """
    current_obj = get(isamAppliance)
    current_lang = current_obj['data']['id']

    if force is True or current_lang != id:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Change the log file language",
                "/isam/applang/v1",
                {'id': id}
            )

    return isamAppliance.create_return_object()


set = update



import logging
from ibmsecurity.utilities import tools
from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/push-notification"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of push notification registrations
    """
    return isamAppliance.invoke_get("Retrieve a list of push notification registrations",
                                    f"{uri}",
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, app_id, check_mode=False, force=False):
    """
    Retrieve a push notification registration
    """
    ret_obj = search(isamAppliance, app_id, check_mode=False, force=False)
    id = ret_obj['data']

    if id == {}:
        logger.info(f"Push notification registration '{app_id}' had no match, skipping retrieval.")
        warnings = [f"Push notification registration '{app_id}' had no match, skipping retrieval."]
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        return _get(isamAppliance, id)


def add(isamAppliance, app_id, platform, provider, check_mode=False, force=False):
    """
    Create a push notification registration
    """
    ret_obj = search(isamAppliance, app_id, check_mode=False, force=False)
    id = ret_obj['data']

    if id != {}:
        logger.info(f"Push notification registration '{app_id}' already exists.  Skipping add.")

    if force is True or id == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_post(
                "Create a push notification registration",
                f"{uri}",
                {
                    'app_id': app_id,
                    'platform': platform,
                    'provider': provider
                },
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def update(isamAppliance, app_id, platform, provider, new_app_id=None, check_mode=False, force=False):
    """
    Update a push notification registration
    """
    ret_obj = search(isamAppliance, app_id=app_id)
    id = ret_obj['data']

    update_required = False

    if id != {}:
        ret_obj = _get(isamAppliance, id=id)

        json_data = {
            'platform': platform,
            'provider': provider
        }
        if new_app_id is not None:
            json_data['app_id'] = new_app_id,
        else:
            json_data['app_id'] = app_id

        sorted_json_data = json_sort(json_data)

        logger.debug(f"Sorted input: {sorted_json_data}")

        del ret_obj['data']['push_id']
        sorted_ret_obj = json_sort(ret_obj['data'])

        logger.debug(f"Sorted existing data: {sorted_ret_obj}")

        if sorted_json_data != sorted_ret_obj:
            update_required = True
        else:
            logger.info(
                f"Input is the same as current Push notification registration '{app_id}'.  Skipping update.")
    else:
        logger.info(f"Push notification registration '{app_id}' does not exists.  Skipping update.")
        warnings = [f"Push notification registration '{app_id}' does not exists.  Skipping update."]
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a push notification registration",
                f"{uri}/{id}",
                json_data,
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def set(isamAppliance, app_id, platform, provider, new_app_id=None, check_mode=False,
        force=False):
    """
    Creating or Modifying a Database PIP
    """
    ret_obj = search(isamAppliance, app_id=app_id)
    id = ret_obj['data']

    if id == {}:
        # If no id was found, Force the add
        return add(isamAppliance, app_id=app_id, platform=platform, provider=provider, check_mode=check_mode,
                   force=True)
    else:
        # Update PIP
        return update(isamAppliance, app_id=app_id, platform=platform, provider=provider, new_app_id=new_app_id,
                      check_mode=check_mode, force=force)


def delete(isamAppliance, app_id, check_mode=False, force=False):
    """
    Delete a push notification registration
    """
    ret_obj = search(isamAppliance, app_id, check_mode=False, force=False)
    id = ret_obj['data']

    if force is True or id != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_delete(
                "Delete a push notification registration",
                f"{uri}/{id}",
                requires_modules=requires_modules, requires_version=requires_version
            )

    if id == {}:
        logger.info(f"Push notification registration '{app_id}' does not exists, skipping delete.")

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare access policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['push_id']

    for obj in ret_obj2['data']:
        del obj['push_id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['push_id'])


def search(isamAppliance, app_id, force=False, check_mode=False):
    """
    Retrieve ID for named Push notification registration
    """
    ret_obj = get_all(isamAppliance)

    ret_obj_new = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['app_id'] == app_id:
            logger.info(f"Found Push notification registration '{app_id}' id: '{obj['push_id']}'")
            ret_obj_new['data'] = obj['push_id']

    return ret_obj_new


def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific Push notification registration",
                                    f"{uri}/{id}")



import ibmsecurity.utilities.tools
import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

module_uri = "/isam/felb/configuration/ha"
requires_modules = None
requires_version = None
requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving HA configuration
    """
    return isamAppliance.invoke_get("Retrieving HA configuration", module_uri,
                                    requires_version=requires_version, requires_modules=requires_modules,
                                    requires_model=requires_model)


def disable(isamAppliance, check_mode=False, force=False):
    """
    Disabling HA
    """

    check_value, warnings = _check_enable(isamAppliance)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Disabling HA", module_uri,
                                               requires_version=requires_version,
                                               requires_modules=requires_modules,
                                               requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def enable(isamAppliance, is_primary, interface, remote, port, health_check_interval,
           health_check_timeout, check_mode=False, force=False):
    """
    Enabling HA
    """

    json_data = {
        "is_primary": is_primary,
        "interface": interface,
        "remote": remote,
        "port": port,
        "health_check_interval": health_check_interval,
        "health_check_timeout": health_check_timeout
    }

    check_value, warnings = _check_enable(isamAppliance)

    if force is True or check_value is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post("Enabling HA", module_uri,
                                             json_data,
                                             requires_version=requires_version,
                                             requires_modules=requires_modules,
                                             requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, is_primary, interface, remote, port, health_check_interval,
           health_check_timeout, check_mode=False, force=False):
    """
    Updating HA configuration
    """

    json_data = {
        "is_primary": is_primary,
        "interface": interface,
        "remote": remote,
        "port": port,
        "health_check_interval": health_check_interval,
        "health_check_timeout": health_check_timeout
    }

    # Call to check function to see if configuration already exist
    update_required, warnings = _check_update(isamAppliance, json_data)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnigns=warnings)
        else:
            return isamAppliance.invoke_put("Updating HA configuration", module_uri,
                                            json_data,
                                            requires_modules=requires_modules,
                                            requires_version=requires_version,
                                            requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, is_primary, interface, remote, port, health_check_interval,
        health_check_timeout, check_mode=False, force=False):
    """
    determines if add or update is used.
    """

    is_enabled, warnings = _check_enable(isamAppliance)


    if is_enabled is False:
        return enable(isamAppliance, is_primary=is_primary, interface=interface, remote=remote, port=port,
                      health_check_interval=health_check_interval,
                      health_check_timeout=health_check_timeout, check_mode=check_mode, force=force)
    elif is_enabled is True:
        return update(isamAppliance, is_primary=is_primary, interface=interface, remote=remote, port=port,
                      health_check_interval=health_check_interval,
                      health_check_timeout=health_check_timeout, check_mode=check_mode, force=force)
    else:
        return isamAppliance.create_return_object(warnings=warnings)

def _check_update(isamAppliance, json_data):
    """
    idempotency test for each parameter
    """

    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']

    if 'enabled' in ret_obj['data']:
        if ret_obj['data']['enabled'] is False:
            return False, warnings
        else:
            del ret_obj['data']['enabled']

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(json_data)
    logger.debug(f"Sorted Existing Data:{sorted_ret_obj}")
    logger.debug(f"Sorted Desired  Data:{sorted_json_data}")
    if sorted_ret_obj != sorted_json_data:
        return True, warnings
    else:
        return False, warnings


def _check_enable(isamAppliance):
    """
    Checks delete function for idempotency
    """

    check_obj = get(isamAppliance)
    warnings = check_obj['warnings']

    if 'enabled' in check_obj['data']:
        if check_obj['data']['enabled'] == True:
            return True, warnings
        else:
            return False, warnings
    else:
        return None, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare FELB HA configuration between two appliances
    """

    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/poc"
requires_modules = ["federation", "mga"]
requires_version = "9.0.1.0"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of point of contact profiles
    """
    return isamAppliance.invoke_get("Retrieve a list of point of contact profiles",
                                    "{0}/profiles".format(uri),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific point of contact profile
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    poc_id = ret_obj['data']

    if poc_id == {}:
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, poc_id)


def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific point of contact profile",
                                    "{0}/profiles/{1}".format(uri, id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get_currentID(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current point of contact profile information
    """
    return isamAppliance.invoke_get("Retrieve the current point of contact profile information", uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def set_current(isamAppliance, name, check_mode=False, force=False):
    """
    Update the current point of contact profile information
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    poc_id = ret_obj['data']
    logger.debug("POC ID to be set Current: {0}".format(poc_id))

    if force is False:
        ret_obj = get_currentID(isamAppliance)
        cur_poc_id = ret_obj['data']['currentProfileId']
        logger.debug("Current POC ID is: {0}".format(cur_poc_id))

    if force is True or poc_id != cur_poc_id:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update the current point of contact profile information", uri,
                {
                    "currentProfileId": poc_id
                },
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def set(isamAppliance, name, description=None, authenticateCallbacks=None, signInCallbacks=None, localIdCallbacks=None,
        signOutCallbacks=None, authnPolicyCallbacks=None, check_mode=False, force=False):
    """
    Creating or Modifying a POC Profile
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        logger.debug("Adding a new POC Profile")
        # Force the add - we already know federation does not exist
        return add(isamAppliance, name, description, authenticateCallbacks, signInCallbacks, localIdCallbacks,
                   signOutCallbacks, authnPolicyCallbacks, check_mode, True)
    else:
        logger.debug("Updating existing POC Profile")
        # Update request
        return update(isamAppliance, name, description, authenticateCallbacks, signInCallbacks, localIdCallbacks,
                      signOutCallbacks, authnPolicyCallbacks, check_mode, force)


def add(isamAppliance, name, description=None, authenticateCallbacks=None, signInCallbacks=None, localIdCallbacks=None,
        signOutCallbacks=None, authnPolicyCallbacks=None, check_mode=False, force=False):
    """
    Create a new point of contact profile
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = _create_json(name, description, authenticateCallbacks, signInCallbacks, localIdCallbacks,
                                     signOutCallbacks, authnPolicyCallbacks)

            return isamAppliance.invoke_post(
                "Create a new point of contact profile",
                "{0}/profiles".format(uri), json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, description=None, authenticateCallbacks=None, signInCallbacks=None,
           localIdCallbacks=None, signOutCallbacks=None, authnPolicyCallbacks=None, check_mode=False, force=False):
    """
    Update a specific point of contact profile
    """
    update_required, poc_id = _check(isamAppliance, name, description, authenticateCallbacks, signInCallbacks,
                                     localIdCallbacks, signOutCallbacks, authnPolicyCallbacks)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = _create_json(name, description, authenticateCallbacks, signInCallbacks, localIdCallbacks,
                                     signOutCallbacks, authnPolicyCallbacks)
            return isamAppliance.invoke_put(
                "Update a specific point of contact profile",
                "{0}/profiles/{1}".format(uri, poc_id), json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a point of contact profile
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            poc_id = ret_obj['data']
            return isamAppliance.invoke_delete(
                "Delete a point of contact profile",
                "{0}/profiles/{1}".format(uri, poc_id),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search Point Of Contact Profile by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.debug("Matched on name: {0}, ID: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']

    return return_obj


def _check(isamAppliance, name, description, authenticateCallbacks, signInCallbacks, localIdCallbacks, signOutCallbacks,
           authnPolicyCallbacks):
    """
    Check and return True if update needed
    """
    update_required = False
    poc_id = None
    ret_obj = get(isamAppliance, name)

    if ret_obj['data'] != {}:
        json_data = _create_json(name, description, authenticateCallbacks, signInCallbacks, localIdCallbacks,
                                 signOutCallbacks, authnPolicyCallbacks)
        poc_id = ret_obj['data']['id']
        del ret_obj['data']['id']
        del ret_obj['data']['isReadOnly']
        logger.debug("Current POC Data: {0}".format(ret_obj['data']))
        logger.debug("POC Data to update: {0}".format(json_data))
        import ibmsecurity.utilities.tools
        if ibmsecurity.utilities.tools.json_sort(ret_obj['data']) != ibmsecurity.utilities.tools.json_sort(json_data):
            update_required = True

    return update_required, poc_id


def _create_json(name, description=None, authenticateCallbacks=None, signInCallbacks=None, localIdCallbacks=None,
                 signOutCallbacks=None, authnPolicyCallbacks=None):
    json_data = {
        "name": name
    }
    if description is not None:
        json_data['description'] = description
    if authenticateCallbacks is not None:
        json_data['authenticateCallbacks'] = authenticateCallbacks
    if signInCallbacks is not None:
        json_data['signInCallbacks'] = signInCallbacks
    if localIdCallbacks is not None:
        json_data['localIdCallbacks'] = localIdCallbacks
    if signOutCallbacks is not None:
        json_data['signOutCallbacks'] = signOutCallbacks
    if authnPolicyCallbacks is not None:
        json_data['authnPolicyCallbacks'] = authnPolicyCallbacks

    return json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Point of Contact profiles between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isvgAppliance, check_mode=False, force=False):
    """
    Get current date/time settings
    """
    return isvgAppliance.invoke_get("Retrieving current date and time settings",
                                    "/time_cfg")


def set(isvgAppliance, ntpServers, timeZone="America/New_York", check_mode=False, force=False):
    """
    Update date/time settings (set NTP server and timezone)
    """
    if force is True or _check(isvgAppliance, timeZone, ntpServers) is False:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_put(
                "Setting date/time settings (NTP)",
                "/time_cfg",
                {
                    "dateTime": "0000-00-00 00:00:00",
                    "timeZone": timeZone,
                    "enableNtp": True,
                    "ntpServers": ntpServers
                })

    return isvgAppliance.create_return_object()


def disable(isvgAppliance, check_mode=False, force=False):
    """
    Disable NTP settings, leaves all other settings intact
    """
    if force is False:
        ret_obj = get(isvgAppliance)

    if force is True or ret_obj['data']['ntpConfig']['enableNtp'] == True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_put(
                "Setting date/time settings (NTP)",
                "/time_cfg",
                {
                    'dateTime': ret_obj['data']['dateTime'],
                    'timeZone': ret_obj['data']['timeZone'],
                    'enableNtp': False,
                    'ntpServers': ','.join(ns['ntpServer'] for ns in ret_obj['data']['ntpConfig']['ntpServers'])
                })

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, timeZone=None, ntpServers=None):
    """
    Check if NTP is already set for syncing date/time
    If timezone or ntpservers provided then also check if those values match what is currently set
    """
    ret_obj = get(isvgAppliance)

    if ret_obj['data']['ntpConfig']['enableNtp']:
        logger.info("NTP is already enabled")
        if timeZone is not None and ret_obj['data']['timeZone'] != timeZone:
            logger.info("Existing timeZone is different")
            return False
        if ntpServers != None:
            existing_ntpServers = list()
            for ntps in ret_obj['data']['ntpConfig']['ntpServers']:
                existing_ntpServers.append(ntps['ntpServer'])
            logger.debug(str(sorted(existing_ntpServers)))
            logger.debug(str(sorted(ntpServers.split(','))))
            if sorted(ntpServers.split(',')) != sorted(existing_ntpServers):
                logger.debug("Existing ntpServers are different")
                return False
        return True
    else:
        return False


def compare(isvgAppliance1, isvgAppliance2):
    """
    Compare date/time settings between two appliances
    """
    ret_obj1 = get(isvgAppliance1)
    ret_obj2 = get(isvgAppliance2)

    # Ignore actual date / time on servers - they should be same if synced correctly
    del ret_obj1['data']['dateTime']
    del ret_obj2['data']['dateTime']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['datetime'])



import logging
import ibmsecurity.utilities.tools as _tools

logger = logging.getLogger(__name__)
requires_modules = ["mga", "federation"]
requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Get runtime tuning settings
    """
    return isamAppliance.invoke_get("Retrieving runtime tuning parameters",
                                    "/mga/runtime_tuning/v1",
                                    requires_modules=requires_modules,
                                    requires_model=requires_model)


def set(isamAppliance, option=None, value=None, values=None, ignore_endpoints=True, check_mode=False, force=False):
    """
    Set a runtime tuning parameter
    option and value are mutually exclusive with values
    Set multiple runtime tuning parameters at once (if values is not none)
    """
    warnings = []
    matches, exists = False, False
    if values is not None:
        # Use the multi option.  Ignore option/value in this case
        return _setMultipleValues(isamAppliance, values, ignore_endpoints, check_mode, force)

    if not force:
        matches, exists, warnings = _check(isamAppliance, option, value)

    if not exists:
        warnings.append(f"Tuning Parameter {option} was not found. set() request will attempt anyway.")

    if force is True or matches is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Setting a runtime tuning parameter",
                f"/mga/runtime_tuning/{option}/v1",
                {
                    'value': value
                }, requires_modules=requires_modules, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _setMultipleValues(isamAppliance, values=None,  ignore_endpoints=True, check_mode=False, force=False):
    """
    Ignore_endpoints does not take the endpoint configuration into account.
    This is typically set using the specific endpoint configurations.
    """
    currentRuntimeParameters = get(isamAppliance)
    warnings = []
    logger.debug("Setting multiple values for runtime tuning parameters")
    if ignore_endpoints:
        logger.info("Ignoring endpoint configuration in comparison")
        warnings.append("Ignoring endpoint configuration in comparison")
        currentRuntimeParameters.pop("endpoints", None)
        values.pop("endpoints", None)
    if force or not _tools.json_equals(currentRuntimeParameters, values):
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Setting multiple runtime tuning parameters",
                "/mga/runtime_tuning/v1",
                values,
                requires_modules=requires_modules, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, option, value):
    """
    Check if tuning parameter option exists and matches value
    """
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']

    matches = False
    exists = False
    try:
        if str(ret_obj['data'][option]) == str(value):
            logger.info(f"Tuning parameter {option}/{value} already set.")
            matches = True
        else:
            logger.info(
                f"Tuning parameter {option} value does not match {ret_obj['data'][option]} != {value}.")
        exists = True
    except Exception:
        logger.info("Runtime tuning parameter does not exist")

    return matches, exists, warnings


def reset(isamAppliance, option, check_mode=False, force=False):
    """
    Reset a tuning parameter
    """
    warnings = []
    matches, exists = False, False

    if force is False:
        matches, exists, warnings = _check(isamAppliance, option, None)

    if force is True or exists:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Reset a runtime tuning parameter to default value",
                f"/mga/runtime_tuning/{option}/v1",
                requires_modules=requires_modules, requires_model=requires_model)
    elif exists is False and warnings == []:
        warnings.append(f"Tuning Parameter {option} was not found. Skipping reset() request.")

    return isamAppliance.create_return_object(warnings=warnings)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare runtime tuning parameters between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    # Ignore differences between endpoints or listening interfaces (compared separately)
    del ret_obj1['data']['endpoints']
    del ret_obj2['data']['endpoints']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['endpoints'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/workflowextension"


def get_all(isvgAppliance, check_mode=False, force=False):
    """
    Retrieve all extensions
    """
    return isvgAppliance.invoke_get("Retrieve all extensions", "{0}".format(uri))


def search(isvgAppliance, name, check_mode=False, force=False):
    """
    Search for existing extension.
    Just care for presence of extension, not its actual content.
    """
    ret_obj = get_all(isvgAppliance)
    return_obj = isvgAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found extension {0}".format(name))
            return_obj['data'] = obj
            return_obj['rc'] = 0
            break

    return return_obj


def add(isvgAppliance, name, extension, check_mode=False, force=False):
    """
    Create an new extension
    For now only support adding extension with the 'Provide XML' method
    """
    ret_obj = search(isvgAppliance, name, check_mode=check_mode, force=force)
    warnings = ret_obj['warnings']

    # extension content must be on a single line
    content = open(extension).read().replace('\n', '')

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # Create a simple json with just the extension data
            extension_json = {
                "name": name,
                "servletName": "",
                "servletDescription": "",
                "servletClass": "",
                "urlPattern": "",
                "loadOnStartup": "0",
                "mode":"advanced",
                "extension": content
            }

            return isvgAppliance.invoke_post(
                "Create a new extension", uri, extension_json, warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def delete(isvgAppliance, name, extension, check_mode=False, force=False):
    """
    Delete an existing extension
    """
    ret_obj = search(isvgAppliance, name, check_mode=check_mode, force=force)
    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            if 'uuid' in ret_obj['data']:
                uuid = ret_obj['data']['uuid']
                warnings = ret_obj['warnings']
                return isvgAppliance.invoke_delete(
                    "Delete an existing extension", "{0}/{1}".format(uri, uuid), warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def update(isvgAppliance, name, extension, check_mode=False, force=False):
    """
    Update an existing extension
    For now only support adding extension with the 'Provide XML' method
    """
    ret_obj = search(isvgAppliance, name, check_mode=check_mode, force=force)
    warnings = ret_obj['warnings']

    # extension content must be on a single line
    content = open(extension).read().replace('\n', '')

    needs_update = True
    if ret_obj['data'] != {}:
        if ret_obj['data']['extension'] == content:
            needs_update = False

    if force is True or needs_update is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            if 'uuid' in ret_obj['data']:
                uuid = ret_obj['data']['uuid']
                # Create a simple json with just the extension data
                extension_json = {
                    "name": name,
                    "servletName": "",
                    "servletDescription": "",
                    "servletClass": "",
                    "urlPattern": "",
                    "loadOnStartup": "0",
                    "mode":"advanced",
                    "extension": content
                }

                return isvgAppliance.invoke_put(
                    "Update an existing extension", "{0}/{1}".format(uri,uuid), extension_json, warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def set(isvgAppliance, name, extension, check_mode=False, force=False):
    """
    Creating or Modifying a extension
    """
    if (search(isvgAppliance, name, extension))['data'] == {}:
        # Force the add - we already know object does not exist
        logger.info("Extension {0} had no match, requesting to add new one.".format(extension))
        return add(isvgAppliance, name, extension, check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("Extension {0} exists, requesting to update.".format(extension))
        return update(isvgAppliance, name, extension, check_mode=check_mode, force=force)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/db_object"


def get(isvgAppliance, check_mode=False, force=False):
    """
    Retrieve identity data store configuration
    """
    return isvgAppliance.invoke_get("Retrieve identity data store configuration entries", "{0}".format(uri))


def search(isvgAppliance, name, check_mode=False, force=False):
    """
    Search for existing identity data store configuration.
    """
    ret_obj = get(isvgAppliance)
    return_obj = isvgAppliance.create_return_object()

    for obj in ret_obj['data']:
        if 'name' in obj and obj['name'] == name:
            logger.info("Found db entry: {0}".format(obj['name']))
            return_obj['data'] = obj
            return_obj['rc'] = 0

    return return_obj


def add(isvgAppliance, name, hostName, port, dbName, adminName, adminPwd, userName, userPwd, configuredAs, retryInterval, maximumRetries, existingDB, oracleSvce, oracleLocationName, dropTables, useSSL, altServerNames, altPortNumbers, connTimeOut, maxConn, minConn, reapTime, agedTimeout, unusedTimeout, check_mode=False, force=False):
    """
    Updating identity data store configuration
    """
    if force is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            if existingDB == ['true']:
                # Reconfiguration does not update the database schema. It configures only the Identity Manager database details.
                action = "reconfigure"
            else:
                # Configuration updates the database schema, in addition it configures the Identity Manager database details.
                action = "configure"

            return isvgAppliance.invoke_post(
                "Configure identity data store", "{0}".format(uri),
                    {
                      "name": name,
                      "hostName": hostName,
                      "port": port,
                      "dbName": dbName,
                      "oracleLocationName": oracleLocationName,
                      "adminName": adminName,
                      "adminPwd": adminPwd,
                      "userName": userName,
                      "userPwd": userPwd,
                      "configuredAs": configuredAs,
                      "dropTables": dropTables,
                      "action": action,
                      "useSSL": useSSL,
                      "altServerNames": altServerNames,
                      "altPortNumbers": altPortNumbers,
                      "retryInterval": retryInterval,
                      "maximumRetries": maximumRetries,
                      "connTimeOut": connTimeOut,
                      "maxConn": maxConn,
                      "minConn": minConn,
                      "reapTime": reapTime,
                      "unusedTimeout": unusedTimeout,
                      "agedTimeout": agedTimeout,
                      "_isNew": True,
                      "existingDB": existingDB,
                      "oracleSvce": oracleSvce
                    })

    return isvgAppliance.create_return_object(changed=False)


def delete(isvgAppliance, name, check_mode=False, force=False):
    """
    Un-configure identity data store
    """
    ret_obj = search(isvgAppliance, name)
    warnings = ret_obj['warnings']

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            ret_obj['data'] = ret_obj['data']
            uuid = ret_obj['data']['uuid']
            return isvgAppliance.invoke_delete(
                "Un-configure identity data store", "{0}/{1}".format(uri, uuid), warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def update(isvgAppliance, name, hostName, port, dbName, adminName, adminPwd, userName, userPwd, configuredAs, retryInterval, maximumRetries, existingDB, oracleSvce, oracleLocationName, dropTables, useSSL, altServerNames, altPortNumbers, connTimeOut, maxConn, minConn, reapTime, agedTimeout, unusedTimeout, check_mode=False, force=False):
    """
    Updating identity data store
    """
    ret_obj = get(isvgAppliance)
    warnings = ret_obj['warnings']

    # JSON payload of interest is at first (and only) position of array
    ret_obj['data'] = ret_obj['data'][0]

    uuid = ret_obj['data']['uuid']

    needs_update = False

    # Create a simple json with just the attributes
    json_data = {
        "name": name,
        "uuid": uuid
    }

    if 'action' in ret_obj['data']:
        del ret_obj['data']['action']
    if 'lastmodified' in ret_obj['data']:
        del ret_obj['data']['lastmodified']
    if 'certCheckSum' in ret_obj['data']:
        del ret_obj['data']['certCheckSum']

    # mandatory attributes
    if hostName is not None:
        json_data['hostName'] = hostName
    elif 'hostName' in ret_obj['data']:
        if ret_obj['data']['hostName'] is not None:
            json_data['hostName'] = ret_obj['data']['hostName']
        else:
            del ret_obj['data']['hostName']
    if port is not None:
        json_data['port'] = port
    elif 'port' in ret_obj['data']:
        if ret_obj['data']['port'] is not None:
            json_data['port'] = ret_obj['data']['port']
        else:
            del ret_obj['data']['port']
    if dbName is not None:
        json_data['dbName'] = dbName
    elif 'dbName' in ret_obj['data']:
        if ret_obj['data']['dbName'] is not None:
            json_data['dbName'] = ret_obj['data']['dbName']
        else:
            del ret_obj['data']['dbName']
    if adminName is not None:
        json_data['adminName'] = adminName
    elif 'adminName' in ret_obj['data']:
        if ret_obj['data']['adminName'] is not None:
            json_data['adminName'] = ret_obj['data']['adminName']
        else:
            del ret_obj['data']['adminName']
    if adminPwd is not None:
        json_data['adminPwd'] = adminPwd
    elif 'adminPwd' in ret_obj['data']:
        if ret_obj['data']['adminPwd'] is not None:
            json_data['adminPwd'] = ret_obj['data']['adminPwd']
        else:
            del ret_obj['data']['adminPwd']
    if userName is not None:
        json_data['userName'] = userName
    elif 'userName' in ret_obj['data']:
        if ret_obj['data']['userName'] is not None:
            json_data['userName'] = ret_obj['data']['userName']
        else:
            del ret_obj['data']['userName']
    if userPwd is not None:
        json_data['userPwd'] = userPwd
    elif 'userPwd' in ret_obj['data']:
        if ret_obj['data']['userPwd'] is not None:
            json_data['userPwd'] = ret_obj['data']['userPwd']
        else:
            del ret_obj['data']['userPwd']
    if configuredAs is not None:
        json_data['configuredAs'] = configuredAs
    elif 'configuredAs' in ret_obj['data']:
        if ret_obj['data']['configuredAs'] is not None:
            json_data['configuredAs'] = ret_obj['data']['configuredAs']
        else:
            del ret_obj['data']['configuredAs']
    if retryInterval is not None:
        json_data['retryInterval'] = retryInterval
    elif 'retryInterval' in ret_obj['data']:
        if ret_obj['data']['retryInterval'] is not None:
            json_data['retryInterval'] = ret_obj['data']['retryInterval']
        else:
            del ret_obj['data']['retryInterval']
    if maximumRetries is not None:
        json_data['maximumRetries'] = maximumRetries
    elif 'maximumRetries' in ret_obj['data']:
        if ret_obj['data']['maximumRetries'] is not None:
            json_data['maximumRetries'] = ret_obj['data']['maximumRetries']
        else:
            del ret_obj['data']['maximumRetries']
    if oracleLocationName is not None:
        json_data['oracleLocationName'] = oracleLocationName
    elif 'oracleLocationName' in ret_obj['data']:
        if ret_obj['data']['oracleLocationName'] is not None:
            json_data['oracleLocationName'] = ret_obj['data']['oracleLocationName']
        else:
            del ret_obj['data']['oracleLocationName']
    # optional attributes
    if dropTables is not None:
        json_data['dropTables'] = dropTables
    elif 'dropTables' in ret_obj['data']:
        if ret_obj['data']['dropTables'] is not None:
            json_data['dropTables'] = ret_obj['data']['dropTables']
        else:
            del ret_obj['data']['dropTables']
    if useSSL is not None:
        json_data['useSSL'] = useSSL
    elif 'useSSL' in ret_obj['data']:
        if ret_obj['data']['useSSL'] is not None:
            json_data['useSSL'] = ret_obj['data']['useSSL']
        else:
            del ret_obj['data']['useSSL']
    if altServerNames is not None:
        json_data['altServerNames'] = altServerNames
    elif 'altServerNames' in ret_obj['data']:
        if ret_obj['data']['altServerNames'] is not None:
            json_data['altServerNames'] = ret_obj['data']['altServerNames']
        else:
            del ret_obj['data']['altServerNames']
    if altPortNumbers is not None:
        json_data['altPortNumbers'] = altPortNumbers
    elif 'altPortNumbers' in ret_obj['data']:
        if ret_obj['data']['altPortNumbers'] is not None:
            json_data['altPortNumbers'] = ret_obj['data']['altPortNumbers']
        else:
            del ret_obj['data']['altPortNumbers']
    if connTimeOut is not None:
        json_data['connTimeOut'] = connTimeOut
    elif 'connTimeOut' in ret_obj['data']:
        if ret_obj['data']['connTimeOut'] is not None:
            json_data['connTimeOut'] = ret_obj['data']['connTimeOut']
        else:
            del ret_obj['data']['connTimeOut']
    if maxConn is not None:
        json_data['maxConn'] = maxConn
    elif 'maxConn' in ret_obj['data']:
        if ret_obj['data']['maxConn'] is not None:
            json_data['maxConn'] = ret_obj['data']['maxConn']
        else:
            del ret_obj['data']['maxConn']
    if minConn is not None:
        json_data['minConn'] = minConn
    elif 'minConn' in ret_obj['data']:
        if ret_obj['data']['minConn'] is not None:
            json_data['minConn'] = ret_obj['data']['minConn']
        else:
            del ret_obj['data']['minConn']
    if reapTime is not None:
        json_data['reapTime'] = reapTime
    elif 'reapTime' in ret_obj['data']:
        if ret_obj['data']['reapTime'] is not None:
            json_data['reapTime'] = ret_obj['data']['reapTime']
        else:
            del ret_obj['data']['reapTime']
    if agedTimeout is not None:
        json_data['agedTimeout'] = agedTimeout
    elif 'agedTimeout' in ret_obj['data']:
        if ret_obj['data']['agedTimeout'] is not None:
            json_data['agedTimeout'] = ret_obj['data']['agedTimeout']
        else:
            del ret_obj['data']['agedTimeout']
    if unusedTimeout is not None:
        json_data['unusedTimeout'] = unusedTimeout
    elif 'unusedTimeout' in ret_obj['data']:
        if ret_obj['data']['unusedTimeout'] is not None:
            json_data['unusedTimeout'] = ret_obj['data']['unusedTimeout']
        else:
            del ret_obj['data']['unusedTimeout']

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(json_data)
    logger.debug("Sorted Existing Data:{0}".format(sorted_ret_obj))
    logger.debug("Sorted Desired  Data:{0}".format(sorted_json_data))
    if sorted_ret_obj != sorted_json_data:
        json_data['_isNew'] = False
        json_data['action'] = "reconfigure"
        if oracleSvce is not None:
            json_data['oracleSvce'] = oracleSvce
        if existingDB is not None:
            json_data['existingDB'] = existingDB
        needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_put(
                "Update identity data store", "{0}/{1}".format(uri, uuid),
                json_data, warnings=warnings)

    return isvgAppliance.create_return_object(changed=False)


def set(isvgAppliance, name, hostName, port, dbName, adminName, adminPwd, userName, userPwd, configuredAs, retryInterval, maximumRetries, existingDB, oracleSvce, oracleLocationName, dropTables="false", useSSL=False, altServerNames="", altPortNumbers="", connTimeOut=180, maxConn=30, minConn=5, reapTime=180, agedTimeout=0, unusedTimeout=1800, check_mode=False, force=False):
    """
    Creating or Modifying a identity data store configuration
    """
    if (search(isvgAppliance, name))['data'] == {}:
        # Force the add - we already know object does not exist
        logger.info("Identity data store {0} had no match, requesting to configure.".format(name))
        return add(isvgAppliance, name, hostName, port, dbName, adminName, adminPwd, userName, userPwd, configuredAs, retryInterval, maximumRetries, existingDB, oracleSvce, oracleLocationName, dropTables, useSSL, altServerNames, altPortNumbers, connTimeOut, maxConn, minConn, reapTime, agedTimeout, unusedTimeout, check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("Identity data store {0} exists, requesting to reconfigure.".format(name))
        return update(isvgAppliance, name, hostName, port, dbName, adminName, adminPwd, userName, userPwd, configuredAs, retryInterval, maximumRetries, existingDB, oracleSvce, oracleLocationName, dropTables, useSSL, altServerNames, altPortNumbers, connTimeOut, maxConn, minConn, reapTime, agedTimeout, unusedTimeout, check_mode=check_mode, force=force)



import logging
import time
from ibmsecurity.appliance.ibmappliance import IBMError

logger = logging.getLogger(__name__)


def restart(isamAppliance, check_mode=False, force=False):
    """
    Restart LMI
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_post("Restarting LMI", "/restarts/restart_server", {})


def get(isamAppliance, check_mode=False, force=False):
    """
    Get LMI Status
    """
    # Be sure to ignore server error
    return isamAppliance.invoke_get("Get LMI Status", "/lmi", ignore_error=True)


def await_startup(isamAppliance, wait_time=300, check_freq=5, start_time=None, check_mode=False, force=False):
    """
    Wait for appliance to bootup or LMI to restart
    Checking lmi responding is best option from REST API perspective

    # Frequency (in seconds) when routine will check if server is up
    # check_freq (seconds)

    # Ideally start_time should be taken before restart request is send to LMI
    # start_time (REST API standard)

    # Time to wait for appliance/lmi to respond and have a different start time
    # wait_time (seconds)

    # Note: This function will not work unless first steps are completed.
    """
    # Get the current start_time if not provided
    if start_time is None:
        ret_obj = get(isamAppliance)
        start_time = ret_obj['data'][0]['start_time']

    sec = 0
    warnings = []

    # Now check if it is up and running
    while 1:
        ret_obj = get(isamAppliance)

        if ret_obj['rc'] == 0 and isinstance(ret_obj['data'], list) and len(ret_obj['data']) > 0 and 'start_time' in \
                ret_obj['data'][0] and ret_obj['data'][0]['start_time'] != start_time:
            logger.info("Server is responding and has a different start time!")
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            time.sleep(check_freq)
            sec += check_freq
            logger.debug(
                f"Server is not responding yet. Waited for {sec} secs, next check in {check_freq} secs.")

        if sec >= wait_time:
            warnings.append(f"The LMI restart not detected or completed, exiting... after {sec} seconds")
            break

    return isamAppliance.create_return_object(warnings=warnings)


def restart_and_wait(isamAppliance, wait_time=300, check_freq=5, check_mode=False, force=False):
    ret_obj = get(isamAppliance)
    _start_time = ret_obj['data'][0]['start_time']

    restart(isamAppliance, check_mode, force)

    return await_startup(isamAppliance, wait_time=wait_time, check_freq=check_freq, start_time=_start_time,
                         check_mode=False, force=False)



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/rsyslog_forwarder/source_names/{0}/instances"
requires_modules = None
requires_version = "9.0.2.1"


def get(isamAppliance, source_name='webseal', check_mode=False, force=False):
    """
    Retrieve the instances for a particular remote syslog forwarding source
    """
    return isamAppliance.invoke_get(
        "Retrieve the instances for a particular remote syslog forwarding source", uri.format(source_name),
        requires_modules=requires_modules, requires_version=requires_version)



import logging
import json
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/fido2/registrations"
requires_modules = ["mga"]
requires_version = "9.0.7.0"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of all FIDO2 registrations
    """
    return isamAppliance.invoke_get("Retrieve a list of all FIDO2 registrations", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)

def get(isamAppliance, credentialId, check_mode=False, force=False):
    """
    Retrieve a specific FIDO2 registration
    """
    ret_obj = isamAppliance.invoke_get("Retrieve a specific FIDO2 registration by credentialId",
                                    f"{uri}/{credentialId}", ignore_error=True,
                                    requires_modules=requires_modules, requires_version=requires_version)

    if ret_obj['rc'] == 404:
        logger.info(f"FIDO2 registration for {credentialId} had no match.")
        return isamAppliance.create_return_object()
    return ret_obj

def delete(isamAppliance, username=None, credentialId=None, check_mode=False, force=False):
    """
    Delete all or specific FIDO2 registrations for a user
    """
    if username is None and credentialId is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot delete FIDO2 registration for unknown username or credentialID. Provide one of the required parameters.")

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        if username is not None:
            return isamAppliance.invoke_delete(
                "Delete all FIDO2 registrations for a user",
                f"{uri}/username/{username}",
                requires_modules=requires_modules, requires_version=requires_version)
        elif credentialId is not None:
            return isamAppliance.invoke_delete(
                "Delete a specific FIDO2 registration for a user",
                f"{uri}/credentialId/{credentialId}",
                requires_modules=requires_modules, requires_version=requires_version)



def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of all update history records
    """
    return isamAppliance.invoke_get("Retrieve a list of all update history records",
                                    "/updates/history/records.json")


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve all update history records
    """
    return isamAppliance.invoke_get("Retrieve all update history records",
                                    "/updates/history/")



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all interfaces
    :rtype: (str, dict)
    """
    return isamAppliance.invoke_get("Retrieving all interfaces", "/net/ifaces", requires_model=requires_model)


def get(isamAppliance, uuid, check_mode=False, force=False):
    """
    Retrieving a single interface
    """
    return isamAppliance.invoke_get("Retrieving a single interface", "/net/ifaces/" + uuid,
                                    requires_model=requires_model)


def _get_ipv4_addresses(isamAppliance, label, vlanId=None):
    """
    Retrieving ipv4 addresses for an interface label
    """
    ret_obj = get_all(isamAppliance)
    warnings = ret_obj['warnings']
    if warnings != []:
        if 'Docker' in warnings[0]:
            return [], warnings

    i = None

    for intfc in ret_obj['data']['interfaces']:
        if intfc['label'] == label and intfc['vlanId'] == vlanId:
            i = intfc
            break

    ipv4_addresses = []
    if i is not None:
        for ads in i['ipv4']['addresses']:
            ipv4_addresses.append(ads['address'])

    return ipv4_addresses, warnings


def get_ipv4_addresses(isamAppliance, label, vlanId=None, check_mode=False, force=False):
    """
    Retrieving ipv4 addresses for an interface label
    """
    ret_obj = isamAppliance.create_return_object()
    ret_obj['data'], ret_obj['warnings'] = _get_ipv4_addresses(isamAppliance, label, vlanId)

    return ret_obj


def get_ipv6_addresses(isamAppliance, label, vlanId=None, check_mode=False, force=False):
    """
    Retrieving ipv6 addresses for an interface label
    """
    ret_obj = get_all(isamAppliance)
    warnings = ret_obj['warnings']
    if warnings != []:
        if 'Docker' in warnings[0]:
            ret_obj['data'] = []
            ret_obj['warnings'] = warnings
            return ret_obj

    i = None

    for intfc in ret_obj['data']['interfaces']:
        if intfc['label'] == label and intfc['vlanId'] == vlanId:
            i = intfc
            break

    ipv6_addresses = []
    if i is not None:
        for adds in i['ipv6']['addresses']:
            ipv6_addresses.append(adds['address'])

    ret_obj = isamAppliance.create_return_object()
    ret_obj['data'] = ipv6_addresses

    return ret_obj


def _get_interface(isamAppliance, label, vlanId=None, check_mode=False, force=False):
    """
    Retrieving a uuid for given interface label
    """
    ret_obj = get_all(isamAppliance)
    warnings = ret_obj['warnings']

    if warnings != []:
        if 'Docker' in warnings[0]:
            return None, warnings

    for intfc in ret_obj['data']['interfaces']:
        if intfc['label'] == label and intfc['vlanId'] == vlanId:
            return intfc, warnings

    return None, warnings


def search(isamAppliance, label=None, vlanId=None, address=None, check_mode=False, force=False):
    """
    Search UUID for given label/vlanid
    """
    ret_obj = isamAppliance.create_return_object()

    if label is not None:
        intf, warnings = _get_interface(isamAppliance, label, vlanId, check_mode, force)
        if warnings != []:
            if 'Docker' in warnings[0]:
                ret_obj['data'] = []
                ret_obj['warnings'] = warnings
                return ret_obj
        if intf == []:
            ret_obj['data'] = []
            ret_obj['warnings'] = warnings
            return ret_obj
        else:
            ret_obj['data'] = intf['uuid']
            return ret_obj
    elif address is not None:
        objs = get_all(isamAppliance)
        warnings = objs['warnings']
        if warnings != []:
            if 'Docker' in warnings[0]:
                ret_obj['data'] = []
                ret_obj['warnings'] = warnings
                return ret_obj
        for intfc in objs['data']['interfaces']:
            for adds in intfc['ipv4']['addresses']:
                if adds['address'] == address:
                    ret_obj['data'] = adds['uuid']
                    return ret_obj

    return ret_obj


def _update_interface(isamAppliance, json_data):
    return isamAppliance.invoke_put("Updating a (VLAN) interface", f"/net/ifaces/{json_data['uuid']}",
                                    json_data, requires_model=requires_model)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare interfaces between 2 appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    if ret_obj1['warnings'] != []:
        if "Docker" in ret_obj1['warnings'][0]:
            return isamAppliance1.create_return_object(warnings=ret_obj1['warnings'])

    if ret_obj2['warnings'] != []:
        if "Docker" in ret_obj2['warnings'][0]:
            return isamAppliance1.create_return_object(warnings=ret_obj1['warnings'])

    for intfc in ret_obj1['data']['interfaces']:
        del intfc['uuid']
        del intfc['bondedTo']
        for ipv4 in intfc['ipv4']['addresses']:
            del ipv4['uuid']
            del ipv4['address']
        for ipv6 in intfc['ipv6']['addresses']:
            del ipv6['uuid']
            del ipv6['address']
    for intfc in ret_obj2['data']['interfaces']:
        del intfc['uuid']
        del intfc['bondedTo']
        for ipv4 in intfc['ipv4']['addresses']:
            del ipv4['uuid']
            del ipv4['address']
        for ipv6 in intfc['ipv6']['addresses']:
            del ipv6['uuid']
            del ipv6['address']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2,
                                                    deleted_keys=['uuid', 'bondedTo',
                                                                  'ipv4/addresses/uuid', 'ipv4/addresses/address',
                                                                  'ipv6/addresses/uuid', 'ipv6/addresses/address'])
